from __future__ import annotations

import dataclasses
import functools
import importlib
import inspect
import json
import math
import re
from abc import ABC, abstractmethod
from bdb import BdbQuit
from collections import ChainMap, defaultdict
from contextlib import contextmanager
from copy import deepcopy
from enum import Enum, IntEnum
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.error import HTTPError

import eth_abi
import eth_abi.abi
import eth_abi.grammar
import eth_abi.packed
import eth_account
import eth_account.messages
import eth_utils
from Crypto.Hash import BLAKE2b, keccak
from typing_extensions import (
    Annotated,
    Literal,
    TypedDict,
    get_args,
    get_origin,
    get_type_hints,
)

from woke.utils import StrEnum

from ..utils.keyed_default_dict import KeyedDefaultDict
from . import hardhat_console
from .blocks import ChainBlocks
from .chain_interfaces import (
    AnvilChainInterface,
    ChainInterfaceAbc,
    GanacheChainInterface,
    GethLikeChainInterfaceAbc,
    HardhatChainInterface,
    TxParams,
)
from .globals import (
    chain_interfaces_manager,
    get_config,
    get_coverage_handler,
    get_exception_handler,
)
from .internal import UnknownEvent, read_from_memory
from .json_rpc.communicator import JsonRpcError
from .primitive_types import Length, ValueRange

if TYPE_CHECKING:
    from .transactions import ChainTransactions, TransactionAbc


# selector => (contract_fqn => pytypes_object)
errors: Dict[bytes, Dict[str, Any]] = {}
# selector => (contract_fqn => pytypes_object)
events: Dict[bytes, Dict[str, Any]] = {}
# contract_fqn => contract type
contracts_by_fqn: Dict[str, Any] = {}
# contract_metadata => contract_fqn
contracts_by_metadata: Dict[bytes, str] = {}
# contract_fqn => tuple of linearized base contract fqns
contracts_inheritance: Dict[str, Tuple[str, ...]] = {}
# contract_fqn => set of REVERT opcode PCs belonging to a revert statement
contracts_revert_index: Dict[str, Set[int]] = {}
# list of pairs of (creation code segments, contract_fqn)
# where creation code segments is a tuple of (length, BLAKE2b hash)
creation_code_index: List[Tuple[Tuple[Tuple[int, bytes], ...], str]] = []


eth_account.Account.enable_unaudited_hdwallet_features()


def get_contracts_by_fqn() -> Dict[str, Any]:
    return contracts_by_fqn


class RevertToSnapshotFailedError(Exception):
    pass


class NotConnectedError(Exception):
    pass


class AlreadyConnectedError(Exception):
    pass


class TransactionConfirmationFailedError(Exception):
    pass


class RequestType(StrEnum):
    ACCESS_LIST = "access_list"
    CALL = "call"
    ESTIMATE = "estimate"
    TX = "tx"


def fix_library_abi(args: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    ret = []
    for arg in args:
        arg = deepcopy(arg)
        if arg["type"] == "tuple":
            fix_library_abi(arg["components"])
        elif "internalType" in arg:
            # internalType was added in 0.5.11
            # versions before 0.5.11 are not supported by Woke anyway
            # but we may fetch ABIs generated by an older version of solc through Etherscan
            if arg["internalType"].startswith("contract "):
                # replace contract type with "address" but keep array suffix (if any)
                arg["type"] = arg["type"].replace(
                    arg["type"].split("[", 1)[0], "address"
                )
            elif arg["internalType"].startswith("enum "):
                # replace enum type with "uint8" but keep array suffix (if any)
                arg["type"] = arg["type"].replace(arg["type"].split("[", 1)[0], "uint8")
        ret.append(arg)
    return ret


class Abi:
    @staticmethod
    def _normalize_input(arguments: Iterable) -> List:
        ret = []
        for arg in arguments:
            if isinstance(arg, Address):
                ret.append(str(arg))
            elif isinstance(arg, Account):
                ret.append(str(arg.address))
            elif isinstance(arg, (list, tuple)):
                ret.append(Abi._normalize_input(arg))
            else:
                ret.append(arg)
        return ret

    @staticmethod
    def _normalize_output(types: Sequence[str], arguments: Sequence) -> Tuple:
        ret = []
        assert len(types) == len(arguments)
        for type, arg in zip(types, arguments):
            type = type.strip()
            if type == "address":
                ret.append(Address(arg))
            elif type.endswith("]"):
                args_type = type[: type.rfind("[")]
                assert isinstance(arg, (list, tuple))
                ret.append(Abi._normalize_output([args_type] * len(arg), arg))
            elif type.startswith("(") and type.endswith(")"):
                abi_type = eth_abi.grammar.parse(type)
                assert isinstance(abi_type, eth_abi.grammar.TupleType)
                ret.append(
                    Abi._normalize_output(
                        [c.to_type_str() for c in abi_type.components], arg
                    )
                )
            else:
                ret.append(arg)
        return tuple(ret)

    @classmethod
    def encode(cls, types: Iterable, arguments: Iterable) -> bytes:
        return eth_abi.abi.encode(types, cls._normalize_input(arguments))

    @classmethod
    def encode_packed(cls, types: Iterable, arguments: Iterable) -> bytes:
        return eth_abi.packed.encode_packed(types, cls._normalize_input(arguments))

    @classmethod
    def encode_with_selector(
        cls, selector: bytes, types: Iterable, arguments: Iterable
    ) -> bytes:
        return selector + cls.encode(types, arguments)

    @classmethod
    def encode_with_signature(
        cls, signature: str, types: Iterable, arguments: Iterable
    ) -> bytes:
        selector = keccak.new(data=signature.encode("utf-8"), digest_bits=256).digest()[
            :4
        ]
        return cls.encode_with_selector(selector, types, arguments)

    @classmethod
    def encode_call(cls, func: Callable, arguments: Iterable) -> bytes:
        def get_class_that_defined_method(meth):
            if isinstance(meth, functools.partial):
                return get_class_that_defined_method(meth.func)
            if inspect.ismethod(meth):
                for c in inspect.getmro(meth.__self__.__class__):
                    if meth.__name__ in c.__dict__:
                        return c
                meth = getattr(
                    meth, "__func__", meth
                )  # fallback to __qualname__ parsing
            if inspect.isfunction(meth):
                c = getattr(
                    inspect.getmodule(meth),
                    meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0],
                    None,
                )
                if isinstance(c, type):
                    return c
            return getattr(
                meth, "__objclass__", None
            )  # handle special descriptor objects

        selector = func.selector
        contract = get_class_that_defined_method(func)
        assert selector in contract._abi  # pyright: ignore reportGeneralTypeIssues
        types = [
            eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
            for arg in fix_library_abi(
                contract._abi[selector][  # pyright: ignore reportGeneralTypeIssues
                    "inputs"
                ]
            )
        ]
        return cls.encode_with_selector(selector, types, arguments)

    @classmethod
    def decode(cls, types: Sequence[str], data: bytes) -> Any:
        return cls._normalize_output(types, eth_abi.abi.decode(types, data))


class Wei(int):
    def to_ether(self) -> float:
        return self / 10**18

    def to_gwei(self) -> float:
        return self / 10**9

    @classmethod
    def from_ether(cls, value: Union[int, float]) -> Wei:
        return cls(int(value * 10**18))

    @classmethod
    def from_gwei(cls, value: Union[int, float]) -> Wei:
        return cls(int(value * 10**9))

    @classmethod
    def from_str(cls, value: str) -> Wei:
        count, unit = value.split()
        return cls(eth_utils.currency.to_wei(float(count), unit))


@functools.total_ordering
class Address:
    ZERO: Address
    _private_key: Optional[bytes] = None

    def __init__(self, address: Union[str, int]) -> None:
        if isinstance(address, int):
            self._address = format(address, "#042x")
        elif isinstance(address, str):
            if not address.startswith(("0x", "0X")):
                address = "0x" + address
            if not eth_utils.address.is_address(address):
                raise ValueError(f"{address} is not a valid address")
            self._address = address
        else:
            raise TypeError("Expected a string or int")

    def __str__(self) -> str:
        return self._address

    def __repr__(self) -> str:
        return self._address

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Address):
            return self._address.lower() == other._address.lower()
        elif isinstance(other, str):
            return self._address.lower() == other.lower()
        elif isinstance(other, Account):
            raise TypeError(
                "Cannot compare Address and Account. Use Account.address instead"
            )
        else:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Address):
            return int(self._address, 16) < int(other._address, 16)
        elif isinstance(other, str):
            return int(self._address, 16) < int(other, 16)
        elif isinstance(other, Account):
            raise TypeError(
                "Cannot compare Address and Account. Use Account.address instead"
            )
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self._address.lower())

    def __bytes__(self) -> bytes:
        return bytes.fromhex(self._address[2:])

    @classmethod
    def from_key(cls, private_key: Union[str, int, bytes]) -> Address:
        acc = eth_account.Account.from_key(private_key)
        ret = cls(acc.address)
        ret._private_key = bytes(acc.key)
        return ret

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic: str,
        passphrase: str = "",
        path: str = "m/44'/60'/0'/0/0",
    ) -> Address:
        acc = eth_account.Account.from_mnemonic(mnemonic, passphrase, path)
        ret = cls(acc.address)
        ret._private_key = bytes(acc.key)
        return ret

    @classmethod
    def from_alias(
        cls,
        alias: str,
        password: Optional[str] = None,
        keystore: Optional[PathLike] = None,
    ) -> Address:
        if keystore is None:
            path = Path(get_config().global_data_path) / "keystore"
        else:
            path = Path(keystore)
        if not path.is_dir():
            raise ValueError(f"Keystore path {path} is not a directory")

        path = path / f"{alias}.json"
        if not path.exists():
            raise ValueError(f"Alias {alias} not found in keystore {path}")

        with path.open() as f:
            data = json.load(f)

        if not data["address"].startswith("0x"):
            data["address"] = "0x" + data["address"]

        if password is None:
            import click

            password = click.prompt(
                f"Password for account {alias}", default="", hide_input=True
            )

        key = eth_account.Account.decrypt(data, password)

        ret = cls(data["address"])
        ret._private_key = bytes(key)
        return ret

    @property
    def private_key(self) -> Optional[bytes]:
        return self._private_key


Address.ZERO = Address(0)


@functools.total_ordering
class Account:
    _address: Address
    _chain: Chain

    def __init__(
        self, address: Union[Address, str, int], chain: Optional[Chain] = None
    ) -> None:
        if chain is None:
            import woke.deployment
            import woke.testing

            if (
                woke.deployment.default_chain.connected
                and woke.testing.default_chain.connected
            ):
                raise ValueError(
                    "Both default_chain and woke.deployment.default_chain are connected. Please specify which chain to use."
                )
            if woke.deployment.default_chain.connected:
                chain = woke.deployment.default_chain
            elif woke.testing.default_chain.connected:
                chain = woke.testing.default_chain
            else:
                raise NotConnectedError("default_chain not connected")

        if isinstance(address, Address):
            self._address = address
        else:
            self._address = Address(address)
        self._chain = chain

    def __str__(self) -> str:
        return self._chain._labels.get(self._address, str(self._address))

    __repr__ = __str__

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Account):
            return self._address == other._address and self._chain == other._chain
        elif isinstance(other, Address):
            raise TypeError(
                "Cannot compare Account to Address. Use Account.address == Address"
            )
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Account):
            if self._chain == other._chain:
                return self._address < other._address
            else:
                raise TypeError(
                    "Cannot compare Accounts from different chains. Compare Account.address instead"
                )
        elif isinstance(other, Address):
            raise TypeError(
                "Cannot compare Account to Address. Use Account.address == Address"
            )
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self._address, self._chain))

    @classmethod
    def new(cls, chain: Optional[Chain] = None) -> Account:
        acc = eth_account.Account.create()
        return cls(Address.from_key(acc.key), chain)

    @classmethod
    def from_key(
        cls, private_key: Union[str, int, bytes], chain: Optional[Chain] = None
    ) -> Account:
        return cls(Address.from_key(private_key), chain)

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic: str,
        passphrase: str = "",
        path: str = "m/44'/60'/0'/0/0",
        chain: Optional[Chain] = None,
    ) -> Account:
        return cls(Address.from_mnemonic(mnemonic, passphrase, path), chain)

    @classmethod
    def from_alias(
        cls,
        alias: str,
        password: Optional[str] = None,
        keystore: Optional[PathLike] = None,
        chain: Optional[Chain] = None,
    ) -> Account:
        return cls(Address.from_alias(alias, password, keystore), chain)

    @property
    def private_key(self) -> Optional[bytes]:
        return self._address.private_key

    @property
    def address(self) -> Address:
        return self._address

    @property
    def label(self) -> Optional[str]:
        return self._chain._labels.get(self._address, None)

    @label.setter
    def label(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise TypeError("label must be a string or None")
        if value is None:
            del self._chain._labels[self._address]
        else:
            self._chain._labels[self._address] = value

    @property
    def balance(self) -> Wei:
        return Wei(self._chain.chain_interface.get_balance(str(self._address)))

    @balance.setter
    def balance(self, value: Union[int, str]) -> None:
        if isinstance(value, str):
            value = Wei.from_str(value)

        if not isinstance(value, int):
            raise TypeError("value must be an integer or string")
        if value < 0:
            raise ValueError("value must be non-negative")
        self._chain.chain_interface.set_balance(str(self.address), value)

    @property
    def code(self) -> bytes:
        return self._chain.chain_interface.get_code(str(self._address))

    @code.setter
    def code(self, value: Union[bytes, bytearray]) -> None:
        if not isinstance(value, (bytes, bytearray)):
            raise TypeError("value must be a bytes object")
        self._chain.chain_interface.set_code(str(self.address), value)

    @property
    def chain(self) -> Chain:
        return self._chain

    @property
    def nonce(self) -> int:
        return self._chain.chain_interface.get_transaction_count(str(self._address))

    @nonce.setter
    def nonce(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("value must be an integer")
        if value < 0:
            raise ValueError("value must be non-negative")
        self._chain.chain_interface.set_nonce(str(self.address), value)
        self._chain._update_nonce(self.address, value)

    def _setup_tx_params(
        self,
        request_type: RequestType,
        data: Union[bytes, bytearray],
        value: Union[int, str],
        from_: Optional[Union[Account, Address, str]],
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]],
        gas_price: Optional[Union[int, str]],
        max_fee_per_gas: Optional[Union[int, str]],
        max_priority_fee_per_gas: Optional[Union[int, str]],
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ],
        type: Optional[int],
    ):
        if isinstance(value, str):
            value = Wei.from_str(value)

        params: TxParams = {
            "data": data,
            "value": value,
            "to": str(self._address),
        }
        if from_ is None:
            if request_type == RequestType.CALL:
                from_ = self._chain.default_call_account
            elif request_type == RequestType.TX:
                from_ = self._chain.default_tx_account
            elif request_type == RequestType.ESTIMATE:
                from_ = self._chain.default_estimate_account
            elif request_type == RequestType.ACCESS_LIST:
                from_ = self._chain.default_access_list_account

        if isinstance(from_, Account):
            if from_.chain != self._chain:
                raise ValueError("`from_` account must belong to this chain")
            params["from"] = str(from_.address)
        elif isinstance(from_, (Address, str)):
            params["from"] = str(from_)
        else:
            raise TypeError("`from_` must be an Account, Address, or str")

        if gas_limit == "max":
            params["gas"] = self._chain.block_gas_limit
        elif gas_limit == "auto":
            params["gas"] = "auto"
        elif isinstance(gas_limit, int):
            params["gas"] = gas_limit
        elif gas_limit is None:
            pass
        else:
            raise TypeError("`gas_limit` must be an int, 'max', 'auto', or None")

        if gas_price is not None:
            if isinstance(gas_price, str):
                gas_price = Wei.from_str(gas_price)
            params["gasPrice"] = gas_price

        if max_fee_per_gas is not None:
            if isinstance(max_fee_per_gas, str):
                max_fee_per_gas = Wei.from_str(max_fee_per_gas)
            params["maxFeePerGas"] = max_fee_per_gas

        if max_priority_fee_per_gas is not None:
            if isinstance(max_priority_fee_per_gas, str):
                max_priority_fee_per_gas = Wei.from_str(max_priority_fee_per_gas)
            params["maxPriorityFeePerGas"] = max_priority_fee_per_gas

        if access_list == "auto":
            params["accessList"] = "auto"
        elif access_list is not None:
            # normalize access_list, all keys should be Address
            tmp_access_list = defaultdict(list)
            for k, v in access_list.items():
                if isinstance(k, Account):
                    k = k.address
                elif isinstance(k, str):
                    k = Address(k)
                elif not isinstance(k, Address):
                    raise TypeError("access_list keys must be Account, Address or str")
                tmp_access_list[k].extend(v)
            access_list = tmp_access_list
            params["accessList"] = [
                {"address": str(k), "storageKeys": [hex(i) for i in v]}
                for k, v in access_list.items()
            ]

        if type is not None:
            params["type"] = type

        return params

    def call(
        self,
        data: Union[bytes, bytearray] = b"",
        value: Union[int, str] = 0,
        from_: Optional[Union[Account, Address, str]] = None,
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None,
        gas_price: Optional[Union[int, str]] = None,
        max_fee_per_gas: Optional[Union[int, str]] = None,
        max_priority_fee_per_gas: Optional[Union[int, str]] = None,
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ] = None,
        type: Optional[int] = None,
        block: Union[
            int,
            Literal["latest"],
            Literal["pending"],
            Literal["earliest"],
            Literal["safe"],
            Literal["finalized"],
        ] = "latest",
    ) -> bytearray:
        params = self._setup_tx_params(
            RequestType.CALL,
            data,
            value,
            from_,
            gas_limit,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            access_list,
            type,
        )
        params = self._chain._build_transaction(RequestType.CALL, params, [], None)

        try:
            coverage_handler = get_coverage_handler()
            if coverage_handler is not None and self._chain._debug_trace_call_supported:
                ret = self._chain.chain_interface.debug_trace_call(params, block)
                coverage_handler.add_coverage(params, self._chain, ret)
                output = bytes.fromhex(ret["returnValue"][2:])
                if ret["failed"]:
                    self._chain._process_revert_data(None, output)
                    raise
            else:
                output = self._chain.chain_interface.call(params, block)
        except JsonRpcError as e:
            self._chain._process_call_revert(e)
            raise

        return bytearray(output)

    def estimate(
        self,
        data: Union[bytes, bytearray] = b"",
        value: Union[int, str] = 0,
        from_: Optional[Union[Account, Address, str]] = None,
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None,
        gas_price: Optional[Union[int, str]] = None,
        max_fee_per_gas: Optional[Union[int, str]] = None,
        max_priority_fee_per_gas: Optional[Union[int, str]] = None,
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ] = None,
        type: Optional[int] = None,
        block: Union[
            int,
            Literal["latest"],
            Literal["pending"],
            Literal["earliest"],
            Literal["safe"],
            Literal["finalized"],
        ] = "pending",
    ) -> int:
        params = self._setup_tx_params(
            RequestType.ESTIMATE,
            data,
            value,
            from_,
            gas_limit,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            access_list,
            type,
        )
        params = self._chain._build_transaction(RequestType.CALL, params, [], None)

        try:
            return self._chain.chain_interface.estimate_gas(params, block)
        except JsonRpcError as e:
            self._chain._process_call_revert(e)
            raise

    def access_list(
        self,
        data: Union[bytes, bytearray] = b"",
        value: Union[int, str] = 0,
        from_: Optional[Union[Account, Address, str]] = None,
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None,
        gas_price: Optional[Union[int, str]] = None,
        max_fee_per_gas: Optional[Union[int, str]] = None,
        max_priority_fee_per_gas: Optional[Union[int, str]] = None,
        type: Optional[int] = None,
        block: Union[
            int,
            Literal["latest"],
            Literal["pending"],
            Literal["earliest"],
            Literal["safe"],
            Literal["finalized"],
        ] = "pending",
    ):
        params = self._setup_tx_params(
            RequestType.ACCESS_LIST,
            data,
            value,
            from_,
            gas_limit,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            {},
            type,
        )
        params = self._chain._build_transaction(
            RequestType.ACCESS_LIST, params, [], None
        )

        try:
            response = self._chain.chain_interface.create_access_list(params, block)
            return {
                Address(e["address"]): [int(s, 16) for s in e["storageKeys"]]
                for e in response["accessList"]
            }, int(response["gasUsed"], 16)
        except JsonRpcError as e:
            self._chain._process_call_revert(e)
            raise

    def transact(
        self,
        data: Union[bytes, bytearray] = b"",
        value: Union[int, str] = 0,
        from_: Optional[Union[Account, Address, str]] = None,
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None,
        gas_price: Optional[Union[int, str]] = None,
        max_fee_per_gas: Optional[Union[int, str]] = None,
        max_priority_fee_per_gas: Optional[Union[int, str]] = None,
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ] = None,
        type: Optional[int] = None,
        confirmations: Optional[int] = None,
    ) -> TransactionAbc[bytearray]:
        tx_params = self._setup_tx_params(
            RequestType.TX,
            data,
            value,
            from_,
            gas_limit,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            access_list,
            type,
        )
        tx_params = self._chain._build_transaction(
            RequestType.CALL, tx_params, [], None
        )

        tx_hash = self._chain._send_transaction(tx_params, from_)

        if "type" not in tx_params:
            from .transactions import LegacyTransaction

            tx_type = LegacyTransaction[bytearray]
        elif tx_params["type"] == 1:
            from .transactions import Eip2930Transaction

            tx_type = Eip2930Transaction[bytearray]
        elif tx_params["type"] == 2:
            from .transactions import Eip1559Transaction

            tx_type = Eip1559Transaction[bytearray]
        else:
            raise ValueError(f"Unknown transaction type {tx_params['type']}")

        tx = tx_type(
            tx_hash,
            tx_params,
            None,
            bytearray,
            self.chain,
        )

        coverage_handler = get_coverage_handler()
        if coverage_handler is not None:
            tx._fetch_debug_trace_transaction()
            coverage_handler.add_coverage(
                tx_params,
                self._chain,
                tx._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
            )

        if confirmations != 0:
            tx.wait(confirmations)

            if self._chain.tx_callback is not None:
                self._chain.tx_callback(tx)

            if tx.error is not None:
                raise tx.error

        return tx

    def sign(self, data: bytes) -> bytes:
        """
        Sign raw data according to EIP-191 type 0x45.
        Specifically, sign(keccak256(b"\x19Ethereum Signed Message:\n" + len(data) + data)) is returned.
        """
        if self.private_key is None:
            return self._chain.chain_interface.sign(str(self._address), data)
        else:
            return bytes(
                eth_account.Account.sign_message(
                    eth_account.messages.encode_defunct(data),
                    self.private_key,
                ).signature
            )

    def sign_hash(self, data_hash: bytes) -> bytes:
        """
        Sign any 32B data (typically keccak256 hash) without prepending any prefix (non EIP-191 compliant).
        This is not recommended for most use cases.
        Specifically, sign(data_hash) is returned.
        """
        if self.private_key is None:
            raise NotImplementedError(
                "Signing data hash without prefix (non EIP-191 compliant) is not supported for accounts without supplied private key"
            )
        else:
            return bytes(
                eth_account.Account.signHash(
                    data_hash,
                    self.private_key,
                ).signature
            )

    def _prepare_eip712_dict(
        self, message: Any, domain: Eip712Domain, client_signing: bool
    ) -> Dict[str, Any]:
        def _get_type(t: Type, options: Optional[Dict[str, Any]] = None) -> str:
            if options is None:
                options = {}

            if get_origin(t) is Annotated:
                args = get_args(t)
                opt = {}

                for arg in args[1:]:
                    if isinstance(arg, Length):
                        opt["length"] = arg.length
                    elif isinstance(arg, ValueRange):
                        opt["min"] = arg.min
                        opt["max"] = arg.max

                return _get_type(args[0], opt)
            elif get_origin(t) is list:
                if "length" in options:
                    return f"{_get_type(get_args(t)[0])}[{options['length']}]"
                else:
                    return f"{_get_type(get_args(t)[0])}[]"
            elif t is int:
                if "min" in options and "max" in options:
                    if options["min"] == 0:
                        bits = math.ceil(math.log2(options["max"] + 1))
                        return f"uint{bits}"
                    else:
                        bits = math.ceil(math.log2(options["max"] - options["min"] + 1))
                        return f"int{bits}"
                else:
                    # kind of fallback, but it's better than nothing
                    return "int256"
            elif t is bytes or t is bytearray:
                if "length" in options:
                    return f"bytes{options['length']}"
                else:
                    return "bytes"
            elif t is str:
                return "string"
            elif issubclass(t, Enum):
                return "uint8"
            elif t is bool:
                return "bool"
            elif issubclass(t, (Account, Address)):
                return "address"
            elif dataclasses.is_dataclass(t):
                return getattr(t, "original_name", t.__name__)
            else:
                raise ValueError(f"Unsupported type {t}")

        def _get_types(t: Type, types: Dict[str, List[Dict[str, str]]]) -> None:
            if not dataclasses.is_dataclass(t):
                return

            name = getattr(t, "original_name", t.__name__)
            if name in types:
                return

            fields = []
            hints = get_type_hints(t, include_extras=True)
            for f in dataclasses.fields(t):
                assert f.name in hints
                fields.append(
                    {
                        "name": f.metadata.get("original_name", f.name),
                        "type": _get_type(hints[f.name]),
                    }
                )

            types[name] = fields

            for f in dataclasses.fields(t):
                assert f.name in hints
                field_type = hints[f.name]
                while (
                    get_origin(field_type) is Annotated
                    or get_origin(field_type) is list
                ):
                    field_type = get_args(field_type)[0]
                if dataclasses.is_dataclass(field_type):
                    _get_types(field_type, types)

        def _get_value(value: Any) -> Any:
            if dataclasses.is_dataclass(value):
                ret = {}
                for f in dataclasses.fields(value):
                    name = f.metadata.get("original_name", f.name)
                    ret[name] = _get_value(getattr(value, f.name))
                return ret
            elif isinstance(value, (list, tuple)):
                return [_get_value(v) for v in value]
            elif isinstance(value, Account):
                return str(value.address)
            elif isinstance(value, Address):
                return str(value)
            elif isinstance(value, IntEnum):
                return int(value)
            elif isinstance(value, (bytes, bytearray)):
                if client_signing:
                    return "0x" + value.hex()
                else:
                    return value
            else:
                return value

        types = {}
        _get_types(type(message), types)

        ret = {
            "types": types,
            "domain": {},
            "primaryType": _get_type(type(message)),
            "message": _get_value(message),
        }

        domain_type = []
        if "name" in domain:
            ret["domain"]["name"] = domain["name"]
            domain_type.append({"name": "name", "type": "string"})
        if "version" in domain:
            ret["domain"]["version"] = domain["version"]
            domain_type.append({"name": "version", "type": "string"})
        if "chainId" in domain:
            ret["domain"]["chainId"] = domain["chainId"]
            domain_type.append({"name": "chainId", "type": "uint256"})
        if "verifyingContract" in domain:
            if isinstance(domain["verifyingContract"], Account):
                ret["domain"]["verifyingContract"] = str(
                    domain["verifyingContract"].address
                )
            else:
                ret["domain"]["verifyingContract"] = str(domain["verifyingContract"])
            domain_type.append({"name": "verifyingContract", "type": "address"})
        if "salt" in domain:
            ret["domain"]["salt"] = "0x" + domain["salt"].hex()
            domain_type.append({"name": "salt", "type": "bytes32"})

        ret["types"]["EIP712Domain"] = domain_type

        return ret

    def sign_structured(
        self, message: Any, domain: Optional[Eip712Domain] = None
    ) -> bytes:
        """
        Sign structured data according to EIP-712. Message can be either a raw dictionary as described in the EIP
        (https://eips.ethereum.org/EIPS/eip-712), or any ABI-compatible dataclass.
        """

        client_signing = self.private_key is None

        if isinstance(message, dict):
            if domain is not None:
                raise ValueError(
                    "Domain cannot be specified when message is a dictionary"
                )
        else:
            if domain is None:
                raise ValueError(
                    "Domain must be specified when message is not a dictionary"
                )
            message = self._prepare_eip712_dict(message, domain, client_signing)

        if client_signing:
            return self._chain.chain_interface.sign_typed(str(self._address), message)
        else:
            return bytes(
                eth_account.Account.sign_message(
                    eth_account.messages.encode_structured_data(message),
                    self.private_key,
                ).signature
            )


Eip712Domain = TypedDict(
    "Eip712Domain",
    {
        "name": str,
        "version": str,
        "chainId": int,
        "verifyingContract": Union[Account, Address, str],
        "salt": bytes,
    },
    total=False,
)


def check_connected(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if not args[0].connected:
            raise NotConnectedError("Not connected to a chain")
        return f(*args, **kwargs)

    return wrapper


class Chain(ABC):
    _connected: bool
    _chain_interface: ChainInterfaceAbc
    _accounts: List[Account]
    _accounts_set: Set[Account]  # for faster lookup
    _nonces: KeyedDefaultDict[Address, int]  # pyright: ignore reportGeneralTypeIssues
    _default_call_account: Optional[Account]
    _default_tx_account: Optional[Account]
    _default_estimate_account: Optional[Account]
    _default_access_list_account: Optional[Account]
    _default_tx_type: int
    _default_tx_confirmations: int
    _deployed_libraries: DefaultDict[bytes, List[Library]]
    _single_source_errors: Set[bytes]
    _snapshots: Dict[str, Dict]
    _blocks: ChainBlocks
    _txs: ChainTransactions
    _chain_id: int
    _labels: Dict[Address, str]
    _require_signed_txs: bool
    _fork: Optional[str]
    _debug_trace_call_supported: bool
    _client_version: str

    tx_callback: Optional[Callable[[TransactionAbc], None]]

    @abstractmethod
    def _connect_setup(
        self, min_gas_price: Optional[int], block_base_fee_per_gas: Optional[int]
    ) -> None:
        ...

    @abstractmethod
    def _connect_finalize(self) -> None:
        ...

    @abstractmethod
    def snapshot(self) -> str:
        ...

    @abstractmethod
    def revert(self, snapshot_id: str) -> None:
        ...

    @abstractmethod
    def _build_transaction(
        self,
        request_type: RequestType,
        params: TxParams,
        arguments: Iterable,
        abi: Optional[Dict],
    ) -> TxParams:
        ...

    @abstractmethod
    def _wait_for_transaction(
        self, tx: TransactionAbc, confirmations: Optional[int]
    ) -> None:
        ...

    @abstractmethod
    def _confirm_transaction(self, tx: TxParams) -> None:
        ...

    @property
    @abstractmethod
    def block_gas_limit(self) -> int:
        ...

    @block_gas_limit.setter
    @abstractmethod
    def block_gas_limit(self, value: int) -> None:
        ...

    @property
    @abstractmethod
    def gas_price(self) -> Wei:
        ...

    @gas_price.setter
    @abstractmethod
    def gas_price(self, value: int) -> None:
        ...

    @property
    @abstractmethod
    def max_priority_fee_per_gas(self) -> Wei:
        ...

    @max_priority_fee_per_gas.setter
    @abstractmethod
    def max_priority_fee_per_gas(self, value: int) -> None:
        ...

    @contextmanager  # pyright: ignore reportGeneralTypeIssues
    @abstractmethod
    def connect(
        self,
        uri: Optional[str] = None,
        *,
        accounts: Optional[int] = None,
        chain_id: Optional[int] = None,
        fork: Optional[str] = None,
        hardfork: Optional[str] = None,
        min_gas_price: Optional[Union[int, str]],
        block_base_fee_per_gas: Optional[Union[int, str]],
    ):
        ...

    def __init__(self):
        self._connected = False

    def _connect(
        self,
        uri: Optional[str] = None,
        *,
        accounts: Optional[int],
        chain_id: Optional[int],
        fork: Optional[str],
        hardfork: Optional[str],
        min_gas_price: Optional[Union[int, str]],
        block_base_fee_per_gas: Optional[Union[int, str]],
    ):
        if self._connected:
            raise AlreadyConnectedError("Already connected to a chain")

        if isinstance(min_gas_price, str):
            min_gas_price = Wei.from_str(min_gas_price)
        if isinstance(block_base_fee_per_gas, str):
            block_base_fee_per_gas = Wei.from_str(block_base_fee_per_gas)

        self._chain_interface = chain_interfaces_manager.get_or_create(
            uri, accounts=accounts, chain_id=chain_id, fork=fork, hardfork=hardfork
        )

        try:
            self._connected = True
            self._client_version = self._chain_interface.get_client_version()

            try:
                self._chain_interface.debug_trace_call(
                    {
                        "type": 0,
                    }
                )
                self._debug_trace_call_supported = True
            except (JsonRpcError, HTTPError):
                self._debug_trace_call_supported = False

            self._chain_id = self._chain_interface.get_chain_id()

            # determine the chain hardfork to set the default tx type
            if isinstance(self._chain_interface, AnvilChainInterface):
                hardfork = self._chain_interface.node_info()["hardFork"]
                if hardfork in {
                    "FRONTIER",
                    "HOMESTEAD",
                    "TANGERINE",
                    "SPURIOUS_DRAGON",
                    "BYZANTIUM",
                    "CONSTANTINOPLE",
                    "PETERSBURG",
                    "ISTANBUL",
                    "MUIR_GLACIER",
                }:
                    self._default_tx_type = 0
                elif hardfork == "BERLIN":
                    self._default_tx_type = 1
                else:
                    self._default_tx_type = 2
            elif isinstance(
                self._chain_interface,
                (GethLikeChainInterfaceAbc, HardhatChainInterface),
            ):
                if self._chain_id in {56, 97}:
                    # BSC clients do not fail on the calls below
                    self._default_tx_type = 1
                else:
                    # TODO this is not the correct flow for Hermez:
                    # EIP-1559 txs should be supported
                    # access lists should not be supported
                    # `type` field in txs should not be used
                    # because Hermez does not implement eth_maxPriorityFeePerGas, fallback to legacy txs
                    try:
                        self._chain_interface.call(
                            {
                                "type": 2,
                                "maxPriorityFeePerGas": 0,
                            }
                        )
                        self._default_tx_type = 2
                    except JsonRpcError:
                        try:
                            self._chain_interface.call(
                                {
                                    "type": 1,
                                    "accessList": [],
                                }
                            )
                            self._default_tx_type = 1
                        except JsonRpcError:
                            self._default_tx_type = 0
            elif isinstance(self._chain_interface, GanacheChainInterface):
                self._default_tx_type = 0
            else:
                raise NotImplementedError(
                    f"Unknown chain interface type: {type(self._chain_interface)}"
                )

            if block_base_fee_per_gas is not None and not isinstance(
                self._chain_interface, GanacheChainInterface
            ):
                try:
                    self._chain_interface.set_next_block_base_fee_per_gas(
                        block_base_fee_per_gas
                    )
                except JsonRpcError:
                    pass

            from .transactions import ChainTransactions

            self._txs = ChainTransactions(self)

            self._accounts = [
                Account(acc, self) for acc in self._chain_interface.get_accounts()
            ]
            self._accounts_set = set(self._accounts)
            self._nonces = KeyedDefaultDict(
                lambda addr: self._chain_interface.get_transaction_count(  # pyright: ignore reportGeneralTypeIssues
                    str(addr)
                )
            )
            self._snapshots = {}
            self._deployed_libraries = defaultdict(list)
            self._default_call_account = (
                self._accounts[0] if len(self._accounts) > 0 else None
            )
            self._default_tx_account = None
            self._default_estimate_account = None
            self._default_access_list_account = None
            self._default_tx_confirmations = 1
            self._blocks = ChainBlocks(self)
            self._labels = {}
            self._fork = fork

            self._single_source_errors = {
                selector
                for selector, sources in errors.items()
                if len({source for fqn, source in sources.items()}) == 1
            }

            self.tx_callback = None

            self._connect_setup(min_gas_price, block_base_fee_per_gas)

            yield self
        except Exception as e:
            if not isinstance(e, BdbQuit):
                exception_handler = get_exception_handler()
                if exception_handler is not None:
                    exception_handler(e)
                raise
        finally:
            self._connect_finalize()
            self._connected = False

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def client_version(self) -> str:
        return self._client_version

    @property
    @check_connected
    def chain_interface(self) -> ChainInterfaceAbc:
        return self._chain_interface

    @property
    @check_connected
    def chain_id(self) -> int:
        return self._chain_id

    @property
    @check_connected
    def accounts(self) -> Tuple[Account, ...]:
        return tuple(self._accounts)

    @property
    @check_connected
    def txs(self) -> ChainTransactions:
        return self._txs

    @property
    @check_connected
    def default_call_account(self) -> Optional[Account]:
        return self._default_call_account

    @default_call_account.setter
    @check_connected
    def default_call_account(self, account: Union[Account, Address, str]) -> None:
        if isinstance(account, Account):
            if account.chain != self:
                raise ValueError("Account is not from this chain")
            self._default_call_account = account
        else:
            self._default_call_account = Account(account, self)

    @property
    @check_connected
    def default_tx_account(self) -> Optional[Account]:
        return self._default_tx_account

    @default_tx_account.setter
    @check_connected
    def default_tx_account(self, account: Union[Account, Address, str]) -> None:
        if isinstance(account, Account):
            if account.chain != self:
                raise ValueError("Account is not from this chain")
            self._default_tx_account = account
        else:
            self._default_tx_account = Account(account, self)

    @property
    @check_connected
    def default_estimate_account(self) -> Optional[Account]:
        return self._default_estimate_account

    @default_estimate_account.setter
    @check_connected
    def default_estimate_account(self, account: Union[Account, Address, str]) -> None:
        if isinstance(account, Account):
            if account.chain != self:
                raise ValueError("Account is not from this chain")
            self._default_estimate_account = account
        else:
            self._default_estimate_account = Account(account, self)

    @property
    @check_connected
    def default_access_list_account(self) -> Optional[Account]:
        return self._default_access_list_account

    @default_access_list_account.setter
    @check_connected
    def default_access_list_account(
        self, account: Union[Account, Address, str]
    ) -> None:
        if isinstance(account, Account):
            if account.chain != self:
                raise ValueError("Account is not from this chain")
            self._default_access_list_account = account
        else:
            self._default_access_list_account = Account(account, self)

    @property
    @check_connected
    def coinbase(self) -> Account:
        return Account(self._chain_interface.get_coinbase(), self)

    @coinbase.setter
    @check_connected
    def coinbase(self, value: Union[Account, Address, str]) -> None:
        if isinstance(value, Account):
            if value.chain != self:
                raise ValueError("Account is not from this chain")
            self._chain_interface.set_coinbase(str(value.address))
        else:
            self._chain_interface.set_coinbase(str(value))

    @property
    @check_connected
    def blocks(self) -> ChainBlocks:
        return self._blocks

    @property
    @check_connected
    def require_signed_txs(self) -> bool:
        return self._require_signed_txs

    @require_signed_txs.setter
    @check_connected
    def require_signed_txs(self, value: bool) -> None:
        self._require_signed_txs = value

    @property
    @check_connected
    def default_tx_type(self) -> int:
        return self._default_tx_type

    @default_tx_type.setter
    @check_connected
    def default_tx_type(self, value: int) -> None:
        if value not in {0, 1, 2}:
            raise ValueError("Invalid transaction type")
        self._default_tx_type = value

    @property
    @check_connected
    def default_tx_confirmations(self) -> int:
        return self._default_tx_confirmations

    @default_tx_confirmations.setter
    @check_connected
    def default_tx_confirmations(self, value: int) -> None:
        if value < 0:
            raise ValueError("Invalid transaction confirmations value")
        self._default_tx_confirmations = value

    @contextmanager
    def change_automine(self, automine: bool):
        if not self._connected:
            raise NotConnectedError("Not connected to a chain")
        automine_was = self._chain_interface.get_automine()
        self._chain_interface.set_automine(automine)
        try:
            yield
        except Exception as e:
            if not isinstance(e, BdbQuit):
                exception_handler = get_exception_handler()
                if exception_handler is not None:
                    exception_handler(e)
                raise
        finally:
            self._chain_interface.set_automine(automine_was)

    @property
    @check_connected
    def automine(self) -> bool:
        return self._chain_interface.get_automine()

    @automine.setter
    @check_connected
    def automine(self, value: bool) -> None:
        self._chain_interface.set_automine(value)

    @check_connected
    def set_next_block_base_fee_per_gas(self, value: Union[int, str]) -> None:
        if isinstance(value, str):
            value = Wei.from_str(value)
        self._chain_interface.set_next_block_base_fee_per_gas(value)

    @check_connected
    def set_next_block_timestamp(self, timestamp: int) -> None:
        self._chain_interface.set_next_block_timestamp(timestamp)

    @check_connected
    def set_min_gas_price(self, value: Union[int, str]) -> None:
        if isinstance(value, str):
            value = Wei.from_str(value)
        self._chain_interface.set_min_gas_price(value)
        self.gas_price = value

    @check_connected
    def set_default_accounts(self, account: Union[Account, Address, str]) -> None:
        if isinstance(account, Account):
            if account.chain != self:
                raise ValueError("Account is not from this chain")
        else:
            account = Account(account, self)

        self._default_call_account = account
        self._default_tx_account = account
        self._default_estimate_account = account
        self._default_access_list_account = account

    @check_connected
    def reset(self) -> None:
        self._chain_interface.reset()

    @check_connected
    def update_accounts(self):
        self._accounts = [
            Account(acc, self) for acc in self._chain_interface.get_accounts()
        ]
        self._accounts_set = set(self._accounts)

    @check_connected
    def mine(self, timestamp_change: Optional[Callable[[int], int]] = None) -> None:
        if timestamp_change is not None:
            block_info = self._chain_interface.get_block("latest")
            assert "timestamp" in block_info
            last_timestamp = int(block_info["timestamp"], 16)
            timestamp = timestamp_change(last_timestamp)
        else:
            timestamp = None

        self._chain_interface.mine(timestamp)

    @check_connected
    def mine_many(
        self, num_blocks: int, timestamp_change: Optional[int] = None
    ) -> None:
        self._chain_interface.mine_many(num_blocks, timestamp_change)

    @contextmanager
    def snapshot_and_revert(self):
        snapshot_id = self.snapshot()
        try:
            yield
        except Exception as e:
            if not isinstance(e, BdbQuit):
                exception_handler = get_exception_handler()
                if exception_handler is not None:
                    exception_handler(e)
                raise
        finally:
            self.revert(snapshot_id)

    def _update_nonce(self, address: Address, nonce: int) -> None:
        self._nonces[address] = nonce

    def _convert_to_web3_type(self, value: Any) -> Any:
        if dataclasses.is_dataclass(value):
            return tuple(
                self._convert_to_web3_type(getattr(value, f.name))
                for f in dataclasses.fields(value)
            )
        elif isinstance(value, list):
            return [self._convert_to_web3_type(v) for v in value]
        elif isinstance(value, tuple):
            return tuple(self._convert_to_web3_type(v) for v in value)
        elif isinstance(value, Account):
            if value.chain != self:
                raise ValueError("Account must belong to this chain")
            return str(value.address)
        elif isinstance(value, Address):
            return str(value)
        elif hasattr(value, "selector") and isinstance(value.selector, bytes):
            instance = value.__self__
            return bytes.fromhex(str(instance.address)[2:]) + value.selector
        else:
            return value

    def _parse_console_log_data(self, data: bytes):
        selector = data[:4]

        if selector not in hardhat_console.abis:
            raise ValueError(f"Unknown selector: {selector.hex()}")
        abi = hardhat_console.abis[selector]

        output_types = [
            eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
            for arg in fix_library_abi(abi)
        ]
        decoded_data = list(Abi.decode(output_types, data[4:]))
        for i in range(len(decoded_data)):
            if abi[i]["type"] == "address":
                decoded_data[i] = Account(decoded_data[i], self)

        if len(decoded_data) == 1:
            decoded_data = decoded_data[0]

        return decoded_data

    def _convert_from_web3_type(
        self, tx: Optional[TransactionAbc], value: Any, expected_type: Type
    ) -> Any:
        if isinstance(expected_type, type(None)):
            return None
        elif expected_type is Callable:
            assert isinstance(value, bytes)
            address = Address("0x" + value[:20].hex())
            fqn = get_fqn_from_address(
                address, tx.block.number - 1 if tx is not None else "latest", self
            )
            if fqn not in contracts_by_fqn:
                raise ValueError(f"Unknown contract: {fqn}")

            module_name, attrs = contracts_by_fqn[fqn]
            obj = getattr(importlib.import_module(module_name), attrs[0])
            for attr in attrs[1:]:
                obj = getattr(obj, attr)

            selector = value[20:24]

            for x in dir(obj):
                m = getattr(obj, x)
                if hasattr(m, "selector") and m.selector == selector:
                    return getattr(obj(address, self), x)

            raise ValueError(
                f"Unable to find function with selector {selector.hex()} in contract {fqn}"
            )
        elif get_origin(expected_type) is list:
            return [
                self._convert_from_web3_type(tx, v, get_args(expected_type)[0])
                for v in value
            ]
        elif get_origin(expected_type) is tuple:
            return tuple(
                self._convert_from_web3_type(tx, v, t)
                for v, t in zip(value, get_args(expected_type))
            )
        elif dataclasses.is_dataclass(expected_type):
            assert isinstance(value, tuple)
            resolved_types = get_type_hints(expected_type)
            field_types = [
                resolved_types[field.name]
                for field in dataclasses.fields(expected_type)
                if field.init
            ]
            assert len(value) == len(field_types)
            converted_values = [
                self._convert_from_web3_type(tx, v, t)
                for v, t in zip(value, field_types)
            ]
            return expected_type(*converted_values)
        elif isinstance(expected_type, type):
            if issubclass(expected_type, Contract):
                return expected_type(value, self)
            elif issubclass(expected_type, Account):
                return Account(value, self)
            elif issubclass(expected_type, IntEnum):
                return expected_type(value)
        return value

    def _process_revert_data(
        self,
        tx: Optional[TransactionAbc],
        revert_data: bytes,
    ):
        from .transactions import UnknownTransactionRevertedError

        selector = revert_data[0:4]
        if selector not in errors:
            e = UnknownTransactionRevertedError(revert_data)
            e.tx = tx
            raise e from None

        if selector not in self._single_source_errors:
            if tx is None:
                e = UnknownTransactionRevertedError(revert_data)
                e.tx = tx
                raise e from None

            # ambiguous error, try to find the source contract
            debug_trace = self._chain_interface.debug_trace_transaction(
                tx.tx_hash, {"enableMemory": True}
            )
            try:
                fqn_overrides: ChainMap[Address, Optional[str]] = ChainMap()
                for i in range(tx.tx_index):
                    prev_tx = tx.block.txs[i]
                    prev_tx._fetch_debug_trace_transaction()
                    process_debug_trace_for_fqn_overrides(
                        prev_tx,
                        prev_tx._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
                        fqn_overrides,
                    )
                fqn = process_debug_trace_for_revert(tx, debug_trace, fqn_overrides)
            except ValueError:
                raise UnknownTransactionRevertedError(revert_data) from None
        else:
            fqn = list(errors[selector].keys())[0]

        module_name, attrs = errors[selector][fqn]
        obj = getattr(importlib.import_module(module_name), attrs[0])
        for attr in attrs[1:]:
            obj = getattr(obj, attr)
        abi = obj._abi

        types = [
            eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
            for arg in fix_library_abi(abi["inputs"])
        ]
        decoded = Abi.decode(types, revert_data[4:])
        generated_error = self._convert_from_web3_type(tx, decoded, obj)
        generated_error.tx = tx
        # raise native pytypes exception on transaction revert
        raise generated_error from None

    def _process_events(self, tx: TransactionAbc, logs: List) -> list:
        if len(logs) == 0:
            return []

        generated_events = []

        non_unique = False
        for log in logs:
            if len(log["topics"]) == 0:
                continue

            selector = log["topics"][0]
            if selector.startswith("0x"):
                selector = selector[2:]
            selector = bytes.fromhex(selector.zfill(64))

            if selector not in events:
                continue

            if len({source for source in events[selector].values()}) > 1:
                non_unique = True
                break

        if non_unique:
            debug_trace = self._chain_interface.debug_trace_transaction(
                tx.tx_hash, {"enableMemory": True}
            )
            fqn_overrides: ChainMap[Address, Optional[str]] = ChainMap()
            for i in range(tx.tx_index):
                prev_tx = tx.block.txs[i]
                prev_tx._fetch_debug_trace_transaction()
                process_debug_trace_for_fqn_overrides(
                    prev_tx,
                    prev_tx._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
                    fqn_overrides,
                )
            event_traces = process_debug_trace_for_events(
                tx, debug_trace, fqn_overrides
            )
            assert len(event_traces) == len(logs)
        else:
            event_traces = [(None, None)] * len(logs)

        for log, (traced_selector, fqn) in zip(logs, event_traces):
            topics = [
                bytes.fromhex(t[2:].zfill(64))
                if t.startswith("0x")
                else bytes.fromhex(t.zfill(64))
                for t in log["topics"]
            ]
            data = (
                bytes.fromhex(log["data"][2:])
                if log["data"].startswith("0x")
                else bytes.fromhex(log["data"])
            )

            if len(topics) == 0:
                generated_events.append(UnknownEvent([], data))
                continue

            selector = topics[0]

            if selector not in events:
                generated_events.append(UnknownEvent(topics, data))
                continue

            if len({source for source in events[selector].values()}) > 1:
                assert traced_selector == selector

                if fqn is None:
                    generated_events.append(UnknownEvent(topics, data))
                    continue

                found = False
                for base_fqn in contracts_inheritance[fqn]:
                    if base_fqn in events[selector]:
                        found = True
                        fqn = base_fqn
                        break

                if not found:
                    generated_events.append(UnknownEvent(topics, data))
                    continue
            else:
                fqn = list(events[selector].keys())[0]

            module_name, attrs = events[selector][fqn]
            obj = getattr(importlib.import_module(module_name), attrs[0])
            for attr in attrs[1:]:
                obj = getattr(obj, attr)
            abi = obj._abi

            topic_index = 1
            types = []

            decoded_indexed = []

            for input in fix_library_abi(abi["inputs"]):
                if input["indexed"]:
                    if (
                        input["type"] in {"string", "bytes"}
                        or input["internalType"].startswith("struct ")
                        or input["type"].endswith("]")
                    ):
                        topic_type = "bytes32"
                    else:
                        topic_type = input["type"]
                    topic_data = log["topics"][topic_index]
                    if topic_data.startswith("0x"):
                        topic_data = topic_data[2:]
                    decoded_indexed.append(
                        Abi.decode([topic_type], bytes.fromhex(topic_data.zfill(64)))[0]
                    )
                    topic_index += 1
                else:
                    types.append(eth_utils.abi.collapse_if_tuple(input))
            decoded = list(Abi.decode(types, bytes.fromhex(log["data"][2:])))
            merged = []

            for input in abi["inputs"]:
                if input["indexed"]:
                    merged.append(decoded_indexed.pop(0))
                else:
                    merged.append(decoded.pop(0))

            merged = tuple(merged)
            generated_event = self._convert_from_web3_type(tx, merged, obj)
            generated_events.append(generated_event)

        return generated_events

    def _process_return_data(
        self, tx: Optional[TransactionAbc], output: bytes, abi: Dict, return_type: Type
    ):
        output_types = [
            eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
            for arg in fix_library_abi(abi["outputs"])
        ]
        decoded_data = Abi.decode(output_types, output)
        if isinstance(decoded_data, (list, tuple)) and len(decoded_data) == 1:
            decoded_data = decoded_data[0]
        return self._convert_from_web3_type(tx, decoded_data, return_type)

    def _process_console_logs(self, trace_output: List[Dict[str, Any]]) -> List:
        hardhat_console_address = bytes.fromhex(
            "000000000000000000636F6e736F6c652e6c6f67"
        )
        console_logs = []
        for trace in trace_output:
            if "action" in trace and "to" in trace["action"]:
                if bytes.fromhex(trace["action"]["to"][2:]) == hardhat_console_address:
                    console_logs.append(
                        self._parse_console_log_data(
                            bytes.fromhex(trace["action"]["input"][2:])
                        )
                    )
        return console_logs

    def _process_console_logs_from_debug_trace(
        self, debug_trace: Dict[str, Any]
    ) -> List:
        hardhat_console_address = Address("0x000000000000000000636F6e736F6c652e6c6f67")
        console_logs = []
        for trace in debug_trace["structLogs"]:
            if trace["op"] == "STATICCALL":
                addr = Address(int(trace["stack"][-2], 16))
                if addr == hardhat_console_address:
                    args_offset = int(trace["stack"][-3], 16)
                    args_size = int(trace["stack"][-4], 16)
                    data = bytes(
                        read_from_memory(args_offset, args_size, trace["memory"])
                    )
                    console_logs.append(self._parse_console_log_data(data))

        return console_logs

    def _process_call_revert(self, e: JsonRpcError):
        try:
            # Hermez does not provide revert data for estimate
            if (
                isinstance(
                    self._chain_interface,
                    (AnvilChainInterface, GethLikeChainInterfaceAbc),
                )
                and e.data["code"] == 3
            ):
                revert_data = e.data["data"]
            elif (
                isinstance(self._chain_interface, GanacheChainInterface)
                and e.data["code"] == -32000
            ):
                revert_data = e.data["data"]
            elif (
                isinstance(self._chain_interface, HardhatChainInterface)
                and e.data["code"] == -32603
            ):
                revert_data = e.data["data"]["data"]
            else:
                raise e from None
        except Exception:
            raise e from None

        if revert_data.startswith("0x"):
            revert_data = revert_data[2:]

        self._process_revert_data(None, bytes.fromhex(revert_data))

    def _send_transaction(
        self, tx_params: TxParams, from_: Optional[Union[Account, Address, str]]
    ) -> str:
        assert "from" in tx_params
        assert "nonce" in tx_params

        if self._chain_id in {56, 97}:
            # BSC doesn't support access lists and tx type
            tx_params.pop("type", None)
            tx_params.pop("accessList", None)

        self._confirm_transaction(tx_params)

        if self.require_signed_txs:
            if isinstance(from_, (Account, Address)):
                key = from_.private_key
            elif from_ is None and self._default_tx_account is not None:
                key = self._default_tx_account.private_key
            else:
                key = None

            tx_params["from"] = eth_utils.address.to_checksum_address(tx_params["from"])

            if "to" in tx_params:
                tx_params["to"] = eth_utils.address.to_checksum_address(tx_params["to"])

            if Account(tx_params["from"], self) in self._accounts_set:
                try:
                    tx_hash = self._chain_interface.send_transaction(tx_params)
                except (ValueError, JsonRpcError) as e:
                    try:
                        tx_hash = e.args[0]["data"]["txHash"]
                    except Exception:
                        raise e from None
            elif key is not None:
                signed_tx = bytes(
                    eth_account.Account.sign_transaction(tx_params, key).rawTransaction
                )
                try:
                    tx_hash = self._chain_interface.send_raw_transaction(signed_tx)
                except (ValueError, JsonRpcError) as e:
                    try:
                        tx_hash = e.args[0]["data"]["txHash"]
                    except Exception:
                        raise e from None
            else:
                raise ValueError(
                    f"Private key for account {tx_params['from']} not known and is not owned by the connected client either."
                )
            self._update_nonce(Address(tx_params["from"]), tx_params["nonce"] + 1)
        else:
            if isinstance(self.chain_interface, AnvilChainInterface):
                try:
                    tx_hash = self.chain_interface.send_unsigned_transaction(tx_params)
                except (ValueError, JsonRpcError) as e:
                    try:
                        tx_hash = e.args[0]["data"]["txHash"]
                    except Exception:
                        raise e
                self._update_nonce(Address(tx_params["from"]), tx_params["nonce"] + 1)
            else:
                sender = Account(tx_params["from"], self)

                with _signer_account(sender):
                    try:
                        tx_hash = self._chain_interface.send_transaction(tx_params)
                    except (ValueError, JsonRpcError) as e:
                        try:
                            tx_hash = e.args[0]["data"]["txHash"]
                        except Exception:
                            raise e
                    self._update_nonce(sender.address, tx_params["nonce"] + 1)
        return tx_hash

    @check_connected
    def _call(
        self,
        abi: Optional[Dict],
        arguments: Iterable,
        params: TxParams,
        return_type: Type,
        block: Union[int, str],
    ) -> Any:
        tx_params = self._build_transaction(RequestType.CALL, params, arguments, abi)
        try:
            coverage_handler = get_coverage_handler()
            if coverage_handler is not None and self._debug_trace_call_supported:
                ret = self._chain_interface.debug_trace_call(tx_params, block)
                coverage_handler.add_coverage(tx_params, self, ret)
                output = bytes.fromhex(ret["returnValue"][2:])
                if ret["failed"]:
                    self._process_revert_data(None, output)
                    raise
            else:
                output = self._chain_interface.call(tx_params, block)
        except JsonRpcError as e:
            self._process_call_revert(e)
            raise

        # deploy
        if "to" not in params:
            return bytearray(output)

        assert abi is not None
        return self._process_return_data(None, output, abi, return_type)

    @check_connected
    def _estimate(
        self,
        abi: Optional[Dict],
        arguments: Iterable,
        params: TxParams,
        block: Union[int, str],
    ) -> int:
        tx_params = self._build_transaction(
            RequestType.ESTIMATE, params, arguments, abi
        )
        try:
            return self._chain_interface.estimate_gas(tx_params, block)
        except JsonRpcError as e:
            self._process_call_revert(e)
            raise

    @check_connected
    def _access_list(
        self,
        abi: Optional[Dict],
        arguments: Iterable,
        params: TxParams,
        block: Union[int, str],
    ):
        tx_params = self._build_transaction(
            RequestType.ACCESS_LIST, params, arguments, abi
        )
        try:
            response = self._chain_interface.create_access_list(tx_params, block)
            return {
                Address(e["address"]): [int(s, 16) for s in e["storageKeys"]]
                for e in response["accessList"]
            }, int(response["gasUsed"], 16)
        except JsonRpcError as e:
            self._process_call_revert(e)
            raise

    @check_connected
    def _transact(
        self,
        abi: Optional[Dict],
        arguments: Iterable,
        params: TxParams,
        return_tx: bool,
        return_type: Type,
        confirmations: Optional[int],
        from_: Optional[Union[Account, Address, str]],
    ) -> Any:
        tx_params = self._build_transaction(RequestType.TX, params, arguments, abi)

        tx_hash = self._send_transaction(tx_params, from_)

        if "type" not in tx_params:
            from woke.development.transactions import LegacyTransaction

            tx_type = LegacyTransaction[return_type]
        elif tx_params["type"] == 1:
            from woke.development.transactions import Eip2930Transaction

            tx_type = Eip2930Transaction[return_type]
        elif tx_params["type"] == 2:
            from woke.development.transactions import Eip1559Transaction

            tx_type = Eip1559Transaction[return_type]
        else:
            raise ValueError(f"Unknown transaction type {tx_params['type']}")

        tx = tx_type(
            tx_hash,
            tx_params,
            abi,
            return_type,
            self,
        )

        coverage_handler = get_coverage_handler()
        if coverage_handler is not None:
            tx._fetch_debug_trace_transaction()
            coverage_handler.add_coverage(
                tx_params,
                self,
                tx._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
            )

        if confirmations != 0:
            tx.wait(confirmations)

            if self.tx_callback is not None:
                self.tx_callback(tx)

            if tx.error is not None:
                raise tx.error

        if return_tx:
            return tx

        return tx.return_value


@contextmanager
def _signer_account(sender: Account):
    chain = sender.chain
    chain_interface = chain.chain_interface
    account_created = True
    if sender not in chain.accounts:
        account_created = False
        if isinstance(chain_interface, (AnvilChainInterface, HardhatChainInterface)):
            chain_interface.impersonate_account(str(sender))
        elif isinstance(chain_interface, GanacheChainInterface):
            chain_interface.add_account(str(sender), "")
            chain.update_accounts()
        else:
            raise NotImplementedError()

    try:
        yield
    finally:
        if not account_created and isinstance(
            chain_interface, (AnvilChainInterface, HardhatChainInterface)
        ):
            chain_interface.stop_impersonating_account(str(sender))


def get_fqn_from_creation_code(creation_code: bytes) -> Tuple[str, int]:
    for creation_code_segments, fqn in creation_code_index:

        length, h = creation_code_segments[0]
        if length > len(creation_code):
            continue
        segment_h = BLAKE2b.new(data=creation_code[:length], digest_bits=256).digest()
        if segment_h != h:
            continue

        creation_code = creation_code[length:]
        found = True
        constructor_offset = length

        for length, h in creation_code_segments[1:]:
            if length + 20 > len(creation_code):
                found = False
                break
            creation_code = creation_code[20:]
            segment_h = BLAKE2b.new(
                data=creation_code[:length], digest_bits=256
            ).digest()
            if segment_h != h:
                found = False
                break
            creation_code = creation_code[length:]
            constructor_offset += length + 20

        if found:
            return fqn, constructor_offset

    raise ValueError("Could not find contract definition from creation code")


def get_fqn_from_address(
    addr: Address, block_number: Union[int, str], chain: Chain
) -> Optional[str]:
    code = chain.chain_interface.get_code(str(addr), block_number)
    metadata = code[-53:]
    if metadata in contracts_by_metadata:
        return contracts_by_metadata[metadata]
    else:
        return None


def get_contract_from_fqn(fqn: str):
    return contracts_by_fqn[fqn]


def process_debug_trace_for_fqn_overrides(
    tx: TransactionAbc,
    debug_trace: Dict[str, Any],
    fqn_overrides: ChainMap[Address, Optional[str]],
) -> None:
    if tx.status == 0:
        return

    trace_is_create = [tx.to is None]
    addresses: List[Optional[Address]] = [tx.to.address if tx.to is not None else None]
    fqns: List[Optional[str]] = []

    fqn_overrides.maps.insert(0, {})

    if tx.to is None:
        fqns.append(None)  # contract is not deployed yet
    else:
        if tx.to.address in fqn_overrides:
            fqns.append(fqn_overrides[tx.to.address])
        else:
            fqns.append(
                get_fqn_from_address(tx.to.address, tx.block.number - 1, tx.chain)
            )

    for i, trace in enumerate(debug_trace["structLogs"]):
        if i > 0:
            prev_trace = debug_trace["structLogs"][i - 1]
            if (
                prev_trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}
                and prev_trace["depth"] == trace["depth"]
            ):
                # precompiled contract was called in the previous trace
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                fqn_overrides.maps.pop(0)
                trace_is_create.pop()
                addresses.pop()
                fqns.pop()

        if trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}:
            trace_is_create.append(False)
            addr = Address(int(trace["stack"][-2], 16))
            addresses.append(addr)
            if addr in fqn_overrides:
                fqns.append(fqn_overrides[addr])
            else:
                fqns.append(get_fqn_from_address(addr, tx.block.number - 1, tx.chain))

            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"CREATE", "CREATE2"}:
            offset = int(trace["stack"][-2], 16)
            length = int(trace["stack"][-3], 16)
            creation_code = read_from_memory(offset, length, trace["memory"])

            trace_is_create.append(True)
            addresses.append(None)
            fqns.append(get_fqn_from_creation_code(creation_code)[0])
            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"INVALID", "RETURN", "REVERT", "STOP", "SELFDESTRUCT"}:
            if trace["op"] == "SELFDESTRUCT":
                if addresses[-1] is not None:
                    fqn_overrides.maps[0][addresses[-1]] = None

            if trace["op"] not in {"INVALID", "REVERT"} and len(fqn_overrides.maps) > 1:
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
            fqn_overrides.maps.pop(0)
            addresses.pop()

            if trace_is_create.pop():
                try:
                    addr = Address(
                        int(debug_trace["structLogs"][i + 1]["stack"][-1], 16)
                    )
                    if addr != Address(0):
                        fqn_overrides.maps[0][addr] = fqns[-1]
                except IndexError:
                    pass
            fqns.pop()


def process_debug_trace_for_revert(
    tx: TransactionAbc,
    debug_trace: Dict,
    fqn_overrides: ChainMap[Address, Optional[str]],
) -> str:
    if tx.to is None:
        origin = get_fqn_from_creation_code(tx.data)[0]
    elif tx.to.address in fqn_overrides:
        origin = fqn_overrides[tx.to.address]
    else:
        origin = get_fqn_from_address(tx.to.address, tx.block.number - 1, tx.chain)

    addresses: List[Optional[Address]] = [tx.to.address if tx.to is not None else None]
    fqns: List[Optional[str]] = [origin]
    trace_is_create: List[bool] = [tx.to is None]
    last_revert_origin = None

    for i, trace in enumerate(debug_trace["structLogs"]):
        if i > 0:
            prev_trace = debug_trace["structLogs"][i - 1]
            if (
                prev_trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}
                and prev_trace["depth"] == trace["depth"]
            ):
                # precompiled contract was called in the previous trace
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                fqn_overrides.maps.pop(0)
                addresses.pop()
                fqns.pop()
                trace_is_create.pop()

        if trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}:
            trace_is_create.append(False)
            addr = Address(int(trace["stack"][-2], 16))
            addresses.append(addr)
            if addr in fqn_overrides:
                fqns.append(fqn_overrides[addr])
            else:
                fqns.append(get_fqn_from_address(addr, tx.block.number - 1, tx.chain))

            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"CREATE", "CREATE2"}:
            offset = int(trace["stack"][-2], 16)
            length = int(trace["stack"][-3], 16)
            creation_code = read_from_memory(offset, length, trace["memory"])

            trace_is_create.append(True)
            addresses.append(None)
            fqns.append(get_fqn_from_creation_code(creation_code)[0])
            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"INVALID", "REVERT"}:
            pc = trace["pc"]
            fqn_overrides.maps.pop(0)
            fqn = fqns.pop()
            addresses.pop()
            trace_is_create.pop()

            if (
                trace["op"] == "REVERT"
                and fqn in contracts_revert_index
                and pc in contracts_revert_index[fqn]
            ):
                last_revert_origin = fqn
        elif trace["op"] in {"RETURN", "STOP", "SELFDESTRUCT"}:
            if len(fqn_overrides.maps) > 1:
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
            fqn_overrides.maps.pop(0)
            addresses.pop()

            if trace_is_create.pop():
                try:
                    addr = Address(
                        int(debug_trace["structLogs"][i + 1]["stack"][-1], 16)
                    )
                    if addr != Address(0):
                        fqn_overrides.maps[0][addr] = fqns[-1]
                except IndexError:
                    pass
            fqns.pop()

    if last_revert_origin is None:
        raise ValueError("Could not find revert origin")
    return last_revert_origin


def process_debug_trace_for_events(
    tx: TransactionAbc,
    debug_trace: Dict,
    fqn_overrides: ChainMap[Address, Optional[str]],
) -> List[Tuple[bytes, Optional[str]]]:
    if tx.to is None:
        origin = get_fqn_from_creation_code(tx.data)[0]
    elif tx.to.address in fqn_overrides:
        origin = fqn_overrides[tx.to.address]
    else:
        origin = get_fqn_from_address(tx.to.address, tx.block.number - 1, tx.chain)

    addresses: List[Optional[Address]] = [tx.to.address if tx.to is not None else None]
    fqns: List[Optional[str]] = [origin]
    trace_is_create: List[bool] = [tx.to is None]
    event_fqns = []

    for i, trace in enumerate(debug_trace["structLogs"]):
        if i > 0:
            prev_trace = debug_trace["structLogs"][i - 1]
            if (
                prev_trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}
                and prev_trace["depth"] == trace["depth"]
            ):
                # precompiled contract was called in the previous trace
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                fqn_overrides.maps.pop(0)
                trace_is_create.pop()
                addresses.pop()
                fqns.pop()

        if trace["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}:
            trace_is_create.append(False)
            addr = Address(int(trace["stack"][-2], 16))
            addresses.append(addr)
            if addr in fqn_overrides:
                fqns.append(fqn_overrides[addr])
            else:
                fqns.append(get_fqn_from_address(addr, tx.block.number - 1, tx.chain))

            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"CREATE", "CREATE2"}:
            offset = int(trace["stack"][-2], 16)
            length = int(trace["stack"][-3], 16)
            creation_code = read_from_memory(offset, length, trace["memory"])

            trace_is_create.append(True)
            addresses.append(None)
            fqns.append(get_fqn_from_creation_code(creation_code)[0])
            fqn_overrides.maps.insert(0, {})
        elif trace["op"] in {"INVALID", "RETURN", "REVERT", "STOP", "SELFDESTRUCT"}:
            if trace["op"] not in {"INVALID", "REVERT"} and len(fqn_overrides.maps) > 1:
                fqn_overrides.maps[1].update(fqn_overrides.maps[0])
            fqn_overrides.maps.pop(0)
            addresses.pop()

            if trace_is_create.pop():
                try:
                    addr = Address(
                        int(debug_trace["structLogs"][i + 1]["stack"][-1], 16)
                    )
                    if addr != Address(0):
                        fqn_overrides.maps[0][addr] = fqns[-1]
                except IndexError:
                    pass
            fqns.pop()
        elif trace["op"] in {"LOG1", "LOG2", "LOG3", "LOG4"}:
            selector = trace["stack"][-3]
            if selector.startswith("0x"):
                selector = selector[2:]
            selector = bytes.fromhex(selector.zfill(64))
            event_fqns.append((selector, fqns[-1]))

    return event_fqns


LIBRARY_PLACEHOLDER_REGEX = re.compile(r"__\$[0-9a-fA-F]{34}\$__")


class Contract(Account):
    _abi: Dict[
        Union[bytes, Literal["constructor"], Literal["fallback"], Literal["receive"]],
        Any,
    ]
    _creation_code: str

    def __init__(
        self, addr: Union[Account, Address, str], chain: Optional[Chain] = None
    ):
        if isinstance(addr, Account):
            if chain is None:
                chain = addr.chain
            elif addr.chain != chain:
                raise ValueError("Account and chain must be from the same chain")
            addr = addr.address
        super().__init__(addr, chain)

    def __str__(self):
        return self._chain._labels.get(
            self.address, f"{self.__class__.__name__}({self.address})"
        )

    __repr__ = __str__

    @classmethod
    def _get_creation_code(
        cls, libraries: Dict[bytes, Tuple[Union[Account, Address], str]]
    ) -> bytes:
        creation_code = cls._creation_code
        for match in LIBRARY_PLACEHOLDER_REGEX.finditer(creation_code):
            lib_id = bytes.fromhex(match.group(0)[3:-3])
            assert (
                lib_id in libraries
            ), f"Address of library {libraries[lib_id][1]} required to generate creation code"

            lib = libraries[lib_id][0]
            if isinstance(lib, Account):
                lib_addr = str(lib.address)[2:]
            elif isinstance(lib, Address):
                lib_addr = str(lib)[2:]
            else:
                raise TypeError()

            creation_code = (
                creation_code[: match.start()] + lib_addr + creation_code[match.end() :]
            )
        return bytes.fromhex(creation_code)

    @classmethod
    def _deploy(
        cls,
        request_type: RequestType,
        arguments: Iterable,
        return_tx: bool,
        return_type: Type,
        from_: Optional[Union[Account, Address, str]],
        value: Union[int, str],
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]],
        libraries: Dict[bytes, Tuple[Union[Account, Address, None], str]],
        chain: Optional[Chain],
        gas_price: Optional[Union[int, str]],
        max_fee_per_gas: Optional[Union[int, str]],
        max_priority_fee_per_gas: Optional[Union[int, str]],
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ],
        type: Optional[int],
        block: Optional[Union[int, str]],
        confirmations: Optional[int],
    ) -> Any:
        if chain is None:
            import woke.deployment
            import woke.testing

            if (
                woke.deployment.default_chain.connected
                and woke.testing.default_chain.connected
            ):
                raise ValueError(
                    "Both default_chain and woke.deployment.default_chain are connected. Please specify which chain to use."
                )
            if woke.deployment.default_chain.connected:
                chain = woke.deployment.default_chain
            elif woke.testing.default_chain.connected:
                chain = woke.testing.default_chain
            else:
                raise NotConnectedError("default_chain not connected")

        creation_code = cls._creation_code
        for match in LIBRARY_PLACEHOLDER_REGEX.finditer(creation_code):
            lib_id = bytes.fromhex(match.group(0)[3:-3])
            assert lib_id in libraries

            lib = libraries[lib_id][0]
            if lib is not None:
                if isinstance(lib, Account):
                    lib_addr = str(lib.address)[2:]
                else:
                    lib_addr = str(lib)[2:]
            elif lib_id in chain._deployed_libraries:
                lib_addr = str(chain._deployed_libraries[lib_id][-1].address)[2:]
            else:
                raise ValueError(f"Library {libraries[lib_id][1]} not deployed")

            creation_code = (
                creation_code[: match.start()] + lib_addr + creation_code[match.end() :]
            )

        return cls._execute(
            chain,
            request_type,
            creation_code,
            arguments,
            return_tx,
            return_type,
            from_,
            None,
            value,
            gas_limit,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            access_list,
            type,
            block,
            confirmations,
        )

    @classmethod
    def _execute(
        cls,
        chain: Chain,
        request_type: RequestType,
        data: str,
        arguments: Iterable,
        return_tx: bool,
        return_type: Type,
        from_: Optional[Union[Account, Address, str]],
        to: Optional[Union[Account, Address, str]],
        value: Union[int, str],
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]],
        gas_price: Optional[Union[int, str]],
        max_fee_per_gas: Optional[Union[int, str]],
        max_priority_fee_per_gas: Optional[Union[int, str]],
        access_list: Optional[
            Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]
        ],
        type: Optional[int],
        block: Optional[Union[int, str]],
        confirmations: Optional[int],
    ):
        if request_type == RequestType.TX and block is not None:
            raise ValueError("block cannot be specified for contract transactions")
        if request_type != RequestType.TX and return_tx:
            raise ValueError("return_tx cannot be specified for non-tx requests")
        if request_type != RequestType.TX and confirmations is not None:
            raise ValueError("confirmations cannot be specified for non-tx requests")
        if confirmations == 0 and not return_tx:
            raise ValueError("confirmations=0 is only valid when return_tx=True")
        if request_type == RequestType.ACCESS_LIST and access_list is not None:
            raise ValueError("access_list cannot be specified for access list requests")

        params: TxParams = {}
        if from_ is not None:
            if isinstance(from_, Account):
                if from_.chain != chain:
                    raise ValueError("`from_` account must belong to this chain")
                params["from"] = str(from_.address)
            else:
                params["from"] = str(from_)

        if isinstance(value, str):
            value = Wei.from_str(value)
        params["value"] = value

        if gas_limit == "max":
            params["gas"] = chain.block_gas_limit
        elif gas_limit == "auto":
            params["gas"] = "auto"
        elif isinstance(gas_limit, int):
            params["gas"] = gas_limit
        elif gas_limit is None:
            pass
        else:
            raise TypeError("`gas_limit` must be an int, 'max', 'auto', or None")

        if to is not None:
            if isinstance(to, Account):
                if to.chain != chain:
                    raise ValueError("`to` account must belong to this chain")
                params["to"] = str(to.address)
            else:
                params["to"] = str(to)

        if gas_price is not None:
            if isinstance(gas_price, str):
                gas_price = Wei.from_str(gas_price)
            params["gasPrice"] = gas_price

        if max_fee_per_gas is not None:
            if isinstance(max_fee_per_gas, str):
                max_fee_per_gas = Wei.from_str(max_fee_per_gas)
            params["maxFeePerGas"] = max_fee_per_gas

        if max_priority_fee_per_gas is not None:
            if isinstance(max_priority_fee_per_gas, str):
                max_priority_fee_per_gas = Wei.from_str(max_priority_fee_per_gas)
            params["maxPriorityFeePerGas"] = max_priority_fee_per_gas

        if access_list == "auto":
            params["accessList"] = "auto"
        elif access_list is not None:
            # normalize access_list, all keys should be Address
            tmp_access_list = defaultdict(list)
            for k, v in access_list.items():
                if isinstance(k, Account):
                    k = k.address
                elif isinstance(k, str):
                    k = Address(k)
                elif not isinstance(k, Address):
                    raise TypeError("access_list keys must be Account, Address or str")
                tmp_access_list[k].extend(v)
            access_list = tmp_access_list
            params["accessList"] = [
                {"address": str(k), "storageKeys": [hex(i) for i in v]}
                for k, v in access_list.items()
            ]

        if type is not None:
            params["type"] = type

        params["data"] = bytes.fromhex(data)

        if to is None:
            abi = cls._abi["constructor"] if "constructor" in cls._abi else None
        else:
            abi = cls._abi[params["data"]]

        if request_type == RequestType.TX:
            return chain._transact(
                abi,
                arguments,
                params,
                return_tx,
                return_type,
                confirmations,
                from_,
            )
        elif request_type == RequestType.CALL:
            if block is None:
                block = "latest"
            return chain._call(abi, arguments, params, return_type, block)
        elif request_type == RequestType.ESTIMATE:
            if block is None:
                block = "pending"

            return chain._estimate(abi, arguments, params, block)
        elif request_type == RequestType.ACCESS_LIST:
            if block is None:
                block = "pending"

            return chain._access_list(abi, arguments, params, block)
        else:
            raise ValueError("invalid request type")


class Library(Contract):
    _library_id: bytes

    @classmethod
    def _deploy(
        cls,
        request_type: RequestType,
        arguments: Iterable,
        return_tx: bool,
        return_type: Type,
        from_: Optional[Union[Account, Address, str]],
        value: Union[int, str],
        gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]],
        libraries: Dict[bytes, Tuple[Union[Account, Address, None], str]],
        chain: Optional[Chain],
        gas_price: Optional[Union[int, str]],
        max_fee_per_gas: Optional[Union[int, str]],
        max_priority_fee_per_gas: Optional[Union[int, str]],
        access_list: Optional[Dict[Union[Account, Address, str], List[int]]],
        type: Optional[int],
        block: Optional[Union[int, str]],
        confirmations: Optional[int],
    ) -> Any:
        if chain is None:
            import woke.deployment
            import woke.testing

            if (
                woke.deployment.default_chain.connected
                and woke.testing.default_chain.connected
            ):
                raise ValueError(
                    "Both default_chain and woke.deployment.default_chain are connected. Please specify which chain to use."
                )
            if woke.deployment.default_chain.connected:
                chain = woke.deployment.default_chain
            elif woke.testing.default_chain.connected:
                chain = woke.testing.default_chain
            else:
                raise NotConnectedError("default_chain not connected")

        lib = super()._deploy(
            request_type,
            arguments,
            return_tx,
            return_type,
            from_,
            value,
            gas_limit,
            libraries,
            chain,
            gas_price,
            max_fee_per_gas,
            max_priority_fee_per_gas,
            access_list,
            type,
            block,
            confirmations,
        )
        if confirmations != 0:
            if return_tx:
                chain._deployed_libraries[cls._library_id].append(lib.return_value)
            else:
                chain._deployed_libraries[cls._library_id].append(lib)
        return lib



import json
import logging
import platform
from pathlib import Path
from typing import Any, Dict, List, Optional

from woke.config import WokeConfig

from .abc import ProtocolAbc
from .http import HttpProtocol
from .ipc import IpcProtocol
from .websocket import WebsocketProtocol

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class JsonRpcError(Exception):
    def __init__(self, data: Dict):
        self.data = data


class JsonRpcCommunicator:
    _protocol: ProtocolAbc
    _request_id: int
    _connected: bool

    def __init__(self, config: WokeConfig, uri: str):
        if uri.startswith(("http://", "https://")):
            self._protocol = HttpProtocol(uri, config.testing.timeout)
        elif uri.startswith(("ws://", "wss://")):
            self._protocol = WebsocketProtocol(uri, config.testing.timeout)
        elif Path(uri).is_socket() or platform.system() == "Windows":
            self._protocol = IpcProtocol(uri, config.testing.timeout)
        else:
            raise ValueError(f"Invalid URI: {uri}")

        self._request_id = 0
        self._connected = False

    def __enter__(self):
        self._protocol.__enter__()
        self._connected = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._protocol.__exit__(exc_type, exc_value, traceback)
        self._connected = False

    @property
    def connected(self) -> bool:
        return self._connected

    def send_request(self, method_name: str, params: Optional[List] = None) -> Any:
        post_data = {
            "jsonrpc": "2.0",
            "method": method_name,
            "params": params if params is not None else [],
            "id": self._request_id,
        }
        logger.info(f"Sending request:\n{post_data}")
        self._request_id += 1

        response = self._protocol.send_recv(json.dumps(post_data))
        logger.info(f"Received response:\n{json.dumps(response)}")
        if "error" in response:
            raise JsonRpcError(response["error"])
        return response["result"]



from pydantic import BaseModel, Extra


def _to_camel(s: str) -> str:
    split = s.split("_")
    return split[0].lower() + "".join([w.capitalize() for w in split[1:]])


class LspModel(BaseModel):
    class Config:
        alias_generator = _to_camel
        allow_population_by_field_name = True
        extra = Extra.ignore



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcRevertStatement

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class RevertStatement(StatementAbc):
    """
    !!! warning
        Only matches reverts with user-defined errors:
        ```solidity
        revert InsufficientBalance(want, have);
        ```
        This is an [ExpressionStatement][woke.ast.ir.statement.expression_statement.ExpressionStatement] with a [FunctionCall][woke.ast.ir.expression.function_call.FunctionCall] expression:
        ```solidity
        revert("Insufficient balance");
        ```
    """

    _ast_node: SolcRevertStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _error_call: FunctionCall

    def __init__(
        self, init: IrInitTuple, revert: SolcRevertStatement, parent: SolidityAbc
    ):
        super().__init__(init, revert, parent)
        self._error_call = FunctionCall(init, revert.error_call, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._error_call

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def error_call(self) -> FunctionCall:
        """
        !!! example
            ```solidity
            InsufficientBalance(want, have)
            ```
            in the following revert statement:
            ```solidity
            revert InsufficientBalance(want, have)
            ```

        Returns:
            Expression representing the error call.
        """
        return self._error_call

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return self.error_call.modifies_state



import asyncio
import logging
import os
import platform

import rich.traceback
import rich_click as click
from click.core import Context
from rich.logging import RichHandler

from .accounts import run_accounts
from .compile import run_compile
from .console import console
from .detect import run_detect
from .fuzz import run_fuzz
from .init import run_init
from .lsp import run_lsp
from .run import run_run
from .svm import run_svm
from .test import run_test

if platform.system() != "Windows":
    try:
        from asyncio import (
            ThreadedChildWatcher,  # pyright: ignore reportGeneralTypeIssues
        )
    except ImportError:
        from woke.utils.threaded_child_watcher import ThreadedChildWatcher


@click.group()
@click.option("--debug/--no-debug", default=False)
@click.option(
    "--profile", is_flag=True, default=False, help="Enable profiling using cProfile."
)
@click.version_option(message="%(version)s", package_name="woke")
@click.pass_context
def main(ctx: Context, debug: bool, profile: bool) -> None:
    if profile:
        import atexit
        import cProfile

        pr = cProfile.Profile()
        pr.enable()

        def exit():
            pr.disable()
            pr.dump_stats("woke.prof")

        atexit.register(exit)

    rich.traceback.install(show_locals=debug, suppress=[click], console=console)
    logging.basicConfig(
        format="%(asctime)s %(name)s: %(message)s",
        handlers=[RichHandler(show_time=False, console=console)],
        level=(logging.WARNING if not debug else logging.DEBUG),
    )

    ctx.ensure_object(dict)
    ctx.obj["debug"] = debug

    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
    else:
        asyncio.get_event_loop_policy().set_child_watcher(ThreadedChildWatcher())

    os.environ["PYTHONBREAKPOINT"] = "ipdb.set_trace"


main.add_command(run_accounts)
main.add_command(run_compile)
main.add_command(run_detect)
main.add_command(run_fuzz)
main.add_command(run_init)
main.add_command(run_lsp)
main.add_command(run_run)
main.add_command(run_svm)
main.add_command(run_test)


@main.command(name="config")
@click.pass_context
def config(ctx: Context) -> None:
    """Print loaded config options in JSON format."""
    from woke.config import WokeConfig

    config = WokeConfig()
    config.load_configs()
    console.print_json(str(config))


def woke_solc() -> None:
    import subprocess
    import sys

    from woke.config import WokeConfig
    from woke.core.solidity_version import SolidityVersion
    from woke.svm import SolcVersionManager

    logging.basicConfig(level=logging.CRITICAL)
    config = WokeConfig()
    config.load_configs()
    svm = SolcVersionManager(config)

    version_file_path = config.global_data_path / ".woke_solc_version"
    if not version_file_path.is_file():
        console.print(
            "Target solc version is not configured. Run 'woke svm use' or 'woke svm switch' command."
        )
        sys.exit(1)

    version = SolidityVersion.fromstring(version_file_path.read_text())
    solc_path = svm.get_path(version)

    if not svm.installed(version):
        console.print(f"solc version {version} is not installed.")
        sys.exit(1)

    proc = subprocess.run(
        [str(solc_path)] + sys.argv[1:],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    print(proc.stdout.decode("utf-8"), end="")
    sys.exit(proc.returncode)



from __future__ import annotations

import importlib
from collections import ChainMap
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, cast

import eth_abi
import eth_abi.abi
import eth_abi.exceptions
import eth_utils
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.text import Text
from rich.tree import Tree

from woke.utils import StrEnum

from . import hardhat_console
from .chain_interfaces import TxParams
from .core import (
    Account,
    Address,
    Chain,
    Contract,
    Wei,
    fix_library_abi,
    get_contracts_by_fqn,
    get_fqn_from_address,
    get_fqn_from_creation_code,
    process_debug_trace_for_fqn_overrides,
)
from .internal import read_from_memory
from .utils import get_contract_info_from_explorer

if TYPE_CHECKING:
    from .transactions import TransactionAbc


def get_precompiled_info(
    addr: Address, data: bytes
) -> Tuple[str, Optional[Tuple[Any, ...]]]:
    if addr == Address(1):
        if len(data) != 128:
            return "ecRecover", None
        return "ecRecover", (data[:32], data[32:64], data[64:96], data[96:128])
    elif addr == Address(2):
        return "SHA2-256", (data,)
    elif addr == Address(3):
        return "RIPEMD-160", (data,)
    elif addr == Address(4):
        return "identity", (data,)
    elif addr == Address(5):
        if len(data) < 96:
            return "modexp", None
        base_length = int.from_bytes(data[:32], "big")
        exp_length = int.from_bytes(data[32:64], "big")
        mod_length = int.from_bytes(data[64:96], "big")
        return "modexp", (
            base_length,
            exp_length,
            mod_length,
            data[96 : 96 + base_length],
            data[96 + base_length : 96 + base_length + exp_length],
            data[
                96
                + base_length
                + exp_length : 96
                + base_length
                + exp_length
                + mod_length
            ],
        )
    elif addr == Address(6):
        if len(data) != 128:
            return "ecAdd", None
        x1 = int.from_bytes(data[:32], "big")
        y1 = int.from_bytes(data[32:64], "big")
        x2 = int.from_bytes(data[64:96], "big")
        y2 = int.from_bytes(data[96:128], "big")
        return "ecAdd", (x1, y1, x2, y2)
    elif addr == Address(7):
        if len(data) != 96:
            return "ecMul", None
        x1 = int.from_bytes(data[:32], "big")
        y1 = int.from_bytes(data[32:64], "big")
        s = int.from_bytes(data[64:96], "big")
        return "ecMul", (x1, y1, s)
    elif addr == Address(8):
        if len(data) % (6 * 32) != 0:
            return "ecPairing", None
        coords = tuple(
            int.from_bytes(data[i : i + 32], "big") for i in range(0, len(data), 32)
        )
        return "ecPairing", coords
    elif addr == Address(9):
        if len(data) != 4 + 64 + 128 + 16 + 1:
            return "Blake2F", None
        rounds = int.from_bytes(data[:4], "big")
        offset = 4
        h = tuple(
            int.from_bytes(data[i : i + 8], "little")
            for i in range(offset, offset + 8 * 8, 8)
        )
        offset += 8 * 8
        m = tuple(
            int.from_bytes(data[i : i + 8], "little")
            for i in range(offset, offset + 16 * 8, 8)
        )
        offset += 16 * 8
        t = tuple(
            int.from_bytes(data[i : i + 8], "little", signed=True)
            for i in range(offset, offset + 2 * 8, 8)
        )
        offset += 2 * 8
        f = data[offset]
        return "Blake2F", (rounds, h, m, t, f)
    else:
        raise ValueError(f"Unknown precompiled contract address: {addr}")


class CallTraceKind(StrEnum):
    CALL = "CALL"
    DELEGATECALL = "DELEGATECALL"
    STATICCALL = "STATICCALL"
    CALLCODE = "CALLCODE"
    CREATE = "CREATE"
    CREATE2 = "CREATE2"
    INTERNAL = "INTERNAL"  # unused


class CallTrace:
    _contract: Optional[Contract]
    _contract_name: Optional[str]
    _function_name: Optional[str]
    _selector: Optional[bytes]
    _function_is_special: bool
    _arguments: Optional[List]
    _status: bool
    _value: Wei
    _kind: CallTraceKind
    _depth: int
    _chain: Chain
    _subtraces: List[CallTrace]
    _parent: Optional[CallTrace]
    _address: Optional[Address]

    def __init__(
        self,
        contract: Optional[Contract],
        contract_name: Optional[str],
        function_name: Optional[str],
        selector: Optional[bytes],
        address: Optional[Address],
        arguments: Optional[Iterable],
        value: int,
        kind: CallTraceKind,
        depth: int,
        chain: Chain,
        function_is_special: bool = False,
    ):
        self._contract = contract
        self._contract_name = contract_name
        self._function_name = function_name
        self._selector = selector
        self._address = address
        self._arguments = list(arguments) if arguments is not None else None
        self._value = Wei(value)
        self._kind = kind
        self._depth = depth
        self._chain = chain
        self._function_is_special = function_is_special
        self._status = True
        self._subtraces = []
        self._parent = None

    def __str__(self):
        console = Console()
        with console.capture() as capture:
            console.print(self)
        return capture.get()

    __repr__ = __str__

    def __rich__(self):
        tree = Tree(self._get_label())
        self._into_tree(tree)
        return tree

    def _into_tree(self, tree: Tree):
        for subtrace in self._subtraces:
            t = tree.add(subtrace._get_label())
            subtrace._into_tree(t)

    def _get_label(self) -> Text:
        ret = Text()

        label = None
        if self.address is not None:
            label = Account(self.address, self.chain).label

        if label is not None:
            contract_name = label
        elif self.contract_name is not None:
            contract_name = self.contract_name
        else:
            contract_name = f"Unknown({self.address})"

        ret.append_text(
            Text.from_markup(f"[bright_magenta]{contract_name}[/bright_magenta].")
        )

        if self.function_name is not None:
            function_name = self.function_name
        else:
            function_name = "???"

        if self.function_is_special:
            ret.append_text(
                Text.from_markup(f"<[bright_magenta]{function_name}[/bright_magenta]>")
            )
        else:
            ret.append_text(
                Text.from_markup(
                    f"[bright_magenta]{self.function_name}[/bright_magenta]"
                )
            )

        if self.kind != CallTraceKind.INTERNAL:
            if self.arguments is not None:
                ret.append("(")
                for i, arg in enumerate(self.arguments):
                    t = Text(repr(arg))
                    ReprHighlighter().highlight(t)
                    ret.append_text(t)
                    if i < len(self.arguments) - 1:
                        ret.append(", ")
                ret.append(")")
            else:
                ret.append("(???)")

        ret.append_text(
            Text.from_markup(
                f" {'[green]✓[/green]' if self.status else '[red]✗[/red]'}"
            )
        )

        if self.kind != CallTraceKind.CALL:
            ret.append_text(
                Text.from_markup(
                    f" [yellow]\[{self.kind}][/yellow]"  # pyright: ignore reportInvalidStringEscapeSequence
                )
            )

        return ret

    @property
    def subtraces(self) -> Tuple[CallTrace, ...]:
        return tuple(self._subtraces)

    @property
    def contract(self) -> Optional[Contract]:
        return self._contract

    @property
    def contract_name(self) -> Optional[str]:
        return self._contract_name

    @property
    def function_name(self) -> Optional[str]:
        return self._function_name

    @property
    def function_is_special(self) -> bool:
        return self._function_is_special

    @property
    def address(self) -> Optional[Address]:
        return self._address

    @property
    def arguments(self) -> Optional[Tuple]:
        if self._arguments is None:
            return None
        return tuple(self._arguments)

    @property
    def status(self) -> bool:
        return self._status

    @property
    def value(self) -> Wei:
        return self._value

    @property
    def kind(self) -> CallTraceKind:
        return self._kind

    @property
    def depth(self) -> int:
        return self._depth

    @property
    def chain(self) -> Chain:
        return self._chain

    @classmethod
    def from_debug_trace(
        cls,
        tx: TransactionAbc,
        trace: Dict[str, Any],
        tx_params: TxParams,
    ):
        fqn_overrides: ChainMap[Address, Optional[str]] = ChainMap()

        # process fqn_overrides for all txs before this one in the same block
        for i in range(tx.tx_index):
            tx_before = tx.block.txs[i]
            tx_before._fetch_debug_trace_transaction()
            process_debug_trace_for_fqn_overrides(
                tx_before,
                tx_before._debug_trace_transaction,  # pyright: ignore reportGeneralTypeIssues
                fqn_overrides,
            )

        assert len(fqn_overrides.maps) == 1

        if tx.to is None:
            try:
                origin_fqn, _ = get_fqn_from_creation_code(tx.data)
            except ValueError:
                origin_fqn = None
        else:
            if tx.to.address in fqn_overrides:
                origin_fqn = fqn_overrides[tx.to.address]
            else:
                origin_fqn = get_fqn_from_address(
                    tx.to.address, tx.block.number - 1, tx.chain
                )

        contracts = [origin_fqn]
        values = [0 if "value" not in tx_params else tx_params["value"]]

        contracts_by_fqn = get_contracts_by_fqn()

        if "value" not in tx_params:
            value = 0
        else:
            value = tx_params["value"]

        explorer_info = None
        precompiled_info = None
        if (
            origin_fqn is None or origin_fqn not in contracts_by_fqn
        ) and tx.to is not None:
            if Address(0) < tx.to.address <= Address(9):
                precompiled_info = get_precompiled_info(
                    tx.to.address, b"" if "data" not in tx_params else tx_params["data"]
                )
            elif tx.chain._fork is not None:
                explorer_info = get_contract_info_from_explorer(
                    tx.to.address, tx.chain.chain_id
                )

        if (
            (origin_fqn is None or origin_fqn not in contracts_by_fqn)
            and explorer_info is None
            and precompiled_info is None
        ):
            root_trace = CallTrace(
                None,
                None,
                None,
                None,
                None if tx.to is None else tx.to.address,
                [b"" if "data" not in tx_params else tx_params["data"]],
                value,
                CallTraceKind.CALL,
                1,
                tx.chain,
                True,
            )
        elif precompiled_info is not None:
            assert tx.to is not None
            precompiled_name, args = precompiled_info
            root_trace = CallTrace(
                None,
                "<precompiled>",
                precompiled_name,
                None,
                tx.to.address,
                args,
                value,
                CallTraceKind.CALL,
                1,
                tx.chain,
                False,
            )
        else:
            if explorer_info is not None:
                contract_name, contract_abi = explorer_info
                obj = None
            else:
                assert origin_fqn is not None
                contract_name = origin_fqn.split(":")[-1]
                module_name, attrs = contracts_by_fqn[origin_fqn]
                obj = getattr(importlib.import_module(module_name), attrs[0])
                for attr in attrs[1:]:
                    obj = getattr(obj, attr)
                contract_abi = obj._abi

            if tx.to is None:
                if "data" not in tx_params or "constructor" not in contract_abi:
                    args = []
                else:
                    _, constructor_offset = get_fqn_from_creation_code(
                        tx_params["data"]
                    )
                    fn_abi = contract_abi["constructor"]
                    output_types = [
                        eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
                        for arg in fix_library_abi(fn_abi["inputs"])
                    ]
                    try:
                        args = list(
                            eth_abi.abi.decode(
                                output_types, tx_params["data"][constructor_offset:]
                            )
                        )
                    except eth_abi.exceptions.DecodingError:
                        args = None
                root_trace = CallTrace(
                    obj,
                    contract_name,
                    "constructor",
                    None,
                    tx.return_value.address if tx.status == 1 else None,
                    args,
                    value,
                    CallTraceKind.CREATE,
                    1,
                    tx.chain,
                    True,
                )
            elif (
                "data" not in tx_params
                or len(tx_params["data"]) == 0
                and "receive" in contract_abi
            ):
                root_trace = CallTrace(
                    obj,
                    contract_name,
                    "receive",
                    None,
                    tx.to.address,
                    [],
                    value,
                    CallTraceKind.CALL,
                    1,
                    tx.chain,
                    True,
                )
            elif (
                "data" not in tx_params
                or len(tx_params["data"]) < 4
                or tx_params["data"][:4] not in contract_abi
            ):
                if "fallback" in contract_abi and (
                    value == 0
                    or contract_abi["fallback"]["stateMutability"] == "payable"
                ):
                    root_trace = CallTrace(
                        obj,
                        contract_name,
                        "fallback",
                        None,
                        tx.to.address,
                        [b"" if "data" not in tx_params else tx_params["data"]],
                        value,
                        CallTraceKind.CALL,
                        1,
                        tx.chain,
                        True,
                    )
                else:
                    root_trace = CallTrace(
                        obj,
                        contract_name,
                        None,
                        None,
                        tx.to.address,
                        [b"" if "data" not in tx_params else tx_params["data"]],
                        value,
                        CallTraceKind.CALL,
                        1,
                        tx.chain,
                        True,
                    )
            else:
                fn_abi = contract_abi[tx_params["data"][:4]]
                output_types = [
                    eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
                    for arg in fix_library_abi(fn_abi["inputs"])
                ]
                try:
                    decoded_data = list(
                        eth_abi.abi.decode(output_types, tx_params["data"][4:])
                    )
                except eth_abi.exceptions.DecodingError:
                    decoded_data = None
                root_trace = CallTrace(
                    obj,
                    contract_name,
                    fn_abi["name"],
                    tx_params["data"][:4],
                    tx.to.address,
                    decoded_data,
                    value,
                    CallTraceKind.CALL,
                    1,
                    tx.chain,
                )

        current_trace = root_trace

        for i, log in enumerate(trace["structLogs"]):
            assert current_trace is not None
            if current_trace.depth != log["depth"]:
                if trace["structLogs"][i - 1]["op"] in {
                    "CALL",
                    "CALLCODE",
                    "DELEGATECALL",
                    "STATICCALL",
                }:
                    # precompiled contract was called in the previous step
                    assert current_trace is not None
                    current_trace = current_trace._parent
                    fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                    fqn_overrides.maps.pop(0)
                    contracts.pop()
                    values.pop()
                else:
                    assert current_trace.depth == log["depth"]

            if log["op"] in {"CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"}:
                if log["op"] in {"CALL", "CALLCODE"}:
                    value = int(log["stack"][-3], 16)
                    args_offset = int(log["stack"][-4], 16)
                    args_size = int(log["stack"][-5], 16)
                elif log["op"] == "DELEGATECALL":
                    value = values[-1]
                    args_offset = int(log["stack"][-3], 16)
                    args_size = int(log["stack"][-4], 16)
                else:
                    value = 0
                    args_offset = int(log["stack"][-3], 16)
                    args_size = int(log["stack"][-4], 16)

                data = bytes(read_from_memory(args_offset, args_size, log["memory"]))

                assert current_trace is not None

                addr = Address(int(log["stack"][-2], 16))
                if addr in fqn_overrides:
                    fqn = fqn_overrides[addr]
                else:
                    fqn = get_fqn_from_address(addr, tx.block.number - 1, tx.chain)

                explorer_info = None
                precompiled_info = None
                if fqn is None and addr != Address(
                    "0x000000000000000000636F6e736F6c652e6c6f67"
                ):
                    if Address(0) < addr <= Address(9):
                        precompiled_info = get_precompiled_info(addr, data)
                    elif tx.chain._fork is not None:
                        explorer_info = get_contract_info_from_explorer(
                            addr, tx.chain.chain_id
                        )

                if fqn is None and explorer_info is None and precompiled_info is None:
                    if addr == Address("0x000000000000000000636F6e736F6c652e6c6f67"):
                        if data[:4] in hardhat_console.abis:
                            fn_abi = hardhat_console.abis[data[:4]]
                            output_types = [
                                eth_utils.abi.collapse_if_tuple(
                                    cast(Dict[str, Any], arg)
                                )
                                for arg in fix_library_abi(fn_abi)
                            ]
                            try:
                                arguments = list(
                                    eth_abi.abi.decode(output_types, data[4:])
                                )
                            except eth_abi.exceptions.DecodingError:
                                arguments = None
                        else:
                            arguments = [data]

                        call_trace = CallTrace(
                            None,
                            "console",
                            "log",
                            data[:4],
                            addr,
                            arguments,
                            value,
                            log["op"],
                            current_trace.depth + 1,
                            tx.chain,
                        )
                    else:
                        call_trace = CallTrace(
                            None,
                            None,
                            None,
                            None,
                            addr,
                            [data],
                            value,
                            log["op"],
                            current_trace.depth + 1,
                            tx.chain,
                            True,
                        )
                elif precompiled_info is not None:
                    call_trace = CallTrace(
                        None,
                        "<precompiled>",
                        precompiled_info[0],
                        None,
                        addr,
                        precompiled_info[1],
                        value,
                        log["op"],
                        current_trace.depth + 1,
                        tx.chain,
                        False,
                    )
                else:
                    if explorer_info is not None:
                        contract_name, contract_abi = explorer_info
                        obj = None
                    else:
                        assert fqn is not None
                        contract_name = fqn.split(":")[-1]
                        module_name, attrs = contracts_by_fqn[fqn]
                        obj = getattr(importlib.import_module(module_name), attrs[0])
                        for attr in attrs[1:]:
                            obj = getattr(obj, attr)
                        contract_abi = obj._abi

                    if args_size >= 4:
                        selector = data[:4]
                        if selector in contract_abi:
                            fn_abi = contract_abi[selector]
                            output_types = [
                                eth_utils.abi.collapse_if_tuple(
                                    cast(Dict[str, Any], arg)
                                )
                                for arg in fix_library_abi(fn_abi["inputs"])
                            ]
                            try:
                                arguments = list(
                                    eth_abi.abi.decode(output_types, data[4:])
                                )
                            except eth_abi.exceptions.DecodingError:
                                arguments = None
                            fn_name = fn_abi["name"]
                            is_special = False
                        elif "fallback" in contract_abi and (
                            value == 0
                            or contract_abi["fallback"]["stateMutability"] == "payable"
                        ):
                            selector = None
                            fn_name = "fallback"
                            arguments = [data]
                            is_special = True
                        else:
                            selector = None
                            fn_name = None
                            arguments = [data]
                            is_special = True
                    else:
                        selector = None
                        if args_size == 0 and "receive" in contract_abi:
                            fn_name = "receive"
                            arguments = []
                            is_special = True
                        elif "fallback" in contract_abi and (
                            value == 0
                            or contract_abi["fallback"]["stateMutability"] == "payable"
                        ):
                            fn_name = "fallback"
                            arguments = [data]
                            is_special = True
                        else:
                            fn_name = None
                            arguments = [data]
                            is_special = True

                    call_trace = CallTrace(
                        obj,
                        contract_name,
                        fn_name,
                        selector,
                        addr,
                        arguments,
                        value,
                        log["op"],
                        current_trace.depth + 1,
                        tx.chain,
                        is_special,
                    )

                current_trace._subtraces.append(call_trace)
                call_trace._parent = current_trace
                current_trace = call_trace
                contracts.append(fqn)
                values.append(value)
                fqn_overrides.maps.insert(0, {})
            elif log["op"] in {"INVALID", "RETURN", "REVERT", "STOP", "SELFDESTRUCT"}:
                if log["op"] in {"INVALID", "REVERT"}:
                    status = False
                else:
                    status = True

                assert current_trace is not None
                while current_trace._kind == CallTraceKind.INTERNAL:
                    current_trace._status = status
                    current_trace = current_trace._parent
                    assert current_trace is not None

                assert current_trace is not None
                if (
                    log["op"] not in {"INVALID", "REVERT"}
                    and len(fqn_overrides.maps) > 1
                ):
                    fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                fqn_overrides.maps.pop(0)

                if current_trace.kind in {CallTraceKind.CREATE, CallTraceKind.CREATE2}:
                    try:
                        address = Address(
                            int(trace["structLogs"][i + 1]["stack"][-1], 16)
                        )
                        if address != Address(0):
                            current_trace._address = address
                            fqn_overrides.maps[0][address] = contracts[-1]
                    except IndexError:
                        pass

                current_trace._status = status
                current_trace = current_trace._parent

                contracts.pop()
                values.pop()
            elif log["op"] in {"CREATE", "CREATE2"}:
                value = int(log["stack"][-1], 16)
                offset = int(log["stack"][-2], 16)
                length = int(log["stack"][-3], 16)

                creation_code = read_from_memory(offset, length, log["memory"])
                try:
                    fqn, constructor_offset = get_fqn_from_creation_code(creation_code)

                    contract_name = fqn.split(":")[-1]
                    module_name, attrs = contracts_by_fqn[fqn]
                    obj = getattr(importlib.import_module(module_name), attrs[0])
                    for attr in attrs[1:]:
                        obj = getattr(obj, attr)
                    contract_abi = obj._abi

                    if "constructor" not in contract_abi:
                        args = []
                    else:
                        fn_abi = contract_abi["constructor"]
                        output_types = [
                            eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
                            for arg in fix_library_abi(fn_abi["inputs"])
                        ]
                        try:
                            args = list(
                                eth_abi.abi.decode(
                                    output_types, creation_code[constructor_offset:]
                                )
                            )
                        except eth_abi.exceptions.DecodingError:
                            args = None
                except ValueError:
                    fqn = None
                    obj = None
                    contract_name = None
                    args = []

                assert current_trace is not None
                call_trace = CallTrace(
                    obj,
                    contract_name,
                    "constructor",
                    None,
                    None,  # to be set later
                    args,
                    value,
                    log["op"],
                    current_trace.depth + 1,
                    tx.chain,
                    True,
                )

                current_trace._subtraces.append(call_trace)
                call_trace._parent = current_trace
                current_trace = call_trace
                contracts.append(fqn)
                values.append(value)
                fqn_overrides.maps.insert(0, {})

        return root_trace



from enum import IntEnum
from typing import Any, List, Optional, Union

from woke.utils import StrEnum

from ..common_structures import (
    Command,
    Diagnostic,
    PartialResultParams,
    Position,
    Range,
    TextDocumentIdentifier,
    TextEdit,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
    WorkspaceEdit,
)
from ..context import LspContext
from ..lsp_data_model import LspModel
from ..utils import uri_to_path


class CodeActionKind(StrEnum):
    EMPTY = ""
    QUICKFIX = "quickfix"
    REFACTOR = "refactor"
    REFACTOR_EXTRACT = "refactor.extract"
    REFACTOR_INLINE = "refactor.inline"
    REFACTOR_REWRITE = "refactor.rewrite"
    SOURCE = "source"
    SOURCE_ORGANIZE_IMPORTS = "source.organizeImports"
    SOURCE_FIX_ALL = "source.fixAll"


class CodeActionTriggerKind(IntEnum):
    INVOKED = 1
    AUTOMATIC = 2


class CodeActionOptions(WorkDoneProgressOptions):
    code_action_kinds: Optional[List[CodeActionKind]] = None
    resolve_provider: Optional[bool] = None


class CodeActionContext(LspModel):
    diagnostics: List[Diagnostic]
    only: Optional[List[CodeActionKind]]
    trigger_kind: Optional[CodeActionTriggerKind]


class CodeActionParams(WorkDoneProgressParams, PartialResultParams):
    text_document: TextDocumentIdentifier
    range: Range
    context: CodeActionContext


class CodeActionDisabled(LspModel):
    reason: str


class CodeAction(LspModel):
    title: str
    kind: Optional[CodeActionKind]
    diagnostics: Optional[List[Diagnostic]]
    is_preferred: Optional[bool]
    disabled: Optional[CodeActionDisabled] = None
    edit: Optional[WorkspaceEdit]
    command: Optional[Command] = None
    data: Optional[Any] = None


async def code_action(
    context: LspContext, params: CodeActionParams
) -> Optional[List[Union[Command, CodeAction]]]:
    path = uri_to_path(params.text_document.uri).resolve()
    line_ending = context.parser.get_line_ending(path)
    if line_ending is None:
        line_ending = "\n"

    ret = []

    for diag in params.context.diagnostics:
        try:
            node = await context.parser.get_node_at_position(
                path, diag.range.start.line, diag.range.start.character
            )
        except KeyError:
            continue

        if diag.code == 4937:
            # No visibility specified. Did you intend to add "public"?
            while node is not None:
                if node.type == "function_definition":
                    break
                node = node.parent

            if node is None or any(
                child.type == "visibility" for child in node.children
            ):
                continue

            for child in node.children:
                if child.type == ")":
                    pos = Position(
                        line=child.end_point[0], character=child.end_point[1] // 2
                    )
                    range = Range(start=pos, end=pos)
                    for visibility in ["public", "private", "external", "internal"]:
                        ret.append(
                            CodeAction(
                                title=f"Add {visibility} visibility",
                                kind=CodeActionKind.QUICKFIX,
                                diagnostics=[diag],
                                is_preferred=visibility == "public",
                                edit=WorkspaceEdit(
                                    changes={
                                        params.text_document.uri: [
                                            TextEdit(
                                                range=range, new_text=f" {visibility}"
                                            )
                                        ]
                                    }
                                ),
                            )
                        )
        elif diag.code == 2018:
            # Function state mutability can be restricted to pure/view
            if "pure" in diag.message:
                mutability = "pure"
            elif "view" in diag.message:
                mutability = "view"
            else:
                continue

            while node is not None:
                if node.type == "function_definition":
                    break
                node = node.parent

            if node is None or any(
                child.type == "state_mutability" for child in node.children
            ):
                continue

            children = [
                child for child in node.children if child.type in {"visibility", ")"}
            ]
            if len(children) == 0:
                continue
            if len(children) > 1:
                child = next(c for c in children if c.type == "visibility")
            else:
                child = children[0]

            pos = Position(line=child.end_point[0], character=child.end_point[1] // 2)
            ret.append(
                CodeAction(
                    title=f"Add {mutability} mutability",
                    kind=CodeActionKind.QUICKFIX,
                    diagnostics=[diag],
                    is_preferred=True,
                    edit=WorkspaceEdit(
                        changes={
                            params.text_document.uri: [
                                TextEdit(
                                    range=Range(start=pos, end=pos),
                                    new_text=f" {mutability}",
                                )
                            ]
                        }
                    ),
                )
            )
        elif diag.code == 1878:
            # SPDX license identifier not provided in source file
            for license in [
                "MIT",
                "Apache-2.0",
                "GPL-3.0",
                "BSD-3-Clause",
                "CC0-1.0",
                "UNLICENSED",
            ]:
                ret.append(
                    CodeAction(
                        title=f"Add SPDX license identifier {license}",
                        kind=CodeActionKind.QUICKFIX,
                        diagnostics=[diag],
                        is_preferred=license == "MIT",
                        edit=WorkspaceEdit(
                            changes={
                                params.text_document.uri: [
                                    TextEdit(
                                        range=Range(
                                            start=Position(line=0, character=0),
                                            end=Position(line=0, character=0),
                                        ),
                                        new_text=f"// SPDX-License-Identifier: {license}{line_ending}",
                                    )
                                ]
                            }
                        ),
                    )
                )

    if len(ret) == 0:
        return None
    return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['woke',
 'woke.analysis',
 'woke.analysis.detectors',
 'woke.analysis.detectors.axelar',
 'woke.ast',
 'woke.ast.ir',
 'woke.ast.ir.declaration',
 'woke.ast.ir.expression',
 'woke.ast.ir.meta',
 'woke.ast.ir.statement',
 'woke.ast.ir.type_name',
 'woke.ast.ir.utils',
 'woke.ast.ir.yul',
 'woke.cli',
 'woke.compiler',
 'woke.compiler.solc_frontend',
 'woke.config',
 'woke.contracts',
 'woke.contracts.woke',
 'woke.core',
 'woke.deployment',
 'woke.development',
 'woke.development.json_rpc',
 'woke.lsp',
 'woke.lsp.commands',
 'woke.lsp.features',
 'woke.lsp.utils',
 'woke.regex_parser',
 'woke.svm',
 'woke.templates.scripts',
 'woke.templates.tests',
 'woke.testing',
 'woke.testing.fuzzing',
 'woke.utils']

package_data = \
{'': ['*']}

install_requires = \
['abch_tree_sitter>=1.1.1,<2.0.0',
 'abch_tree_sitter_solidity>=1.2.0,<2.0.0',
 'aiofiles>=0.8,<0.9',
 'aiohttp>=3.8,<4.0',
 'click>=8,<9',
 'eth-abi>=4.0.0b2,<5.0.0',
 'eth-account>=0.8,<0.9',
 'eth-hash[pycryptodome]>=0.5.1,<0.6.0',
 'eth-utils>=2.1,<3.0',
 'graphviz>=0.19,<0.20',
 'intervaltree>=3.1,<4.0',
 'ipdb>=0.13.9,<0.14.0',
 'networkx>=2.5,<3.0',
 'parsimonious>=0.9,<0.10',
 'pathvalidate>=2.5,<3.0',
 'pydantic>=1.9.1,<2.0.0',
 'pytest>=7,<8',
 'rich-click>=1.6,<2.0',
 'rich>=13.3.2,<14.0.0',
 'tblib>=1.7,<2.0',
 'tomli>=2,<3',
 'typing-extensions>=4,<5',
 'watchdog>=2.2.0,<2.3.0',
 'websocket-client>=1.4,<2.0']

extras_require = \
{':sys_platform == "win32"': ['pywin32>=302'],
 'dev': ['black>=22,<23',
         'mkdocs-material>=9,<10',
         'mkdocstrings[python]>=0.19,<0.20',
         'pymdown-extensions>=9,<10',
         'pygments>=2,<3',
         'isort>=5,<6',
         'pillow>=9,<10',
         'cairosvg>=2.7,<3.0'],
 'tests': ['pytest-asyncio>=0.17,<0.18', 'GitPython>=3.1.20,<4.0.0']}

entry_points = \
{'console_scripts': ['woke = woke.cli.__main__:main',
                     'woke-solc = woke.cli.__main__:woke_solc']}

setup_kwargs = {
    'name': 'woke',
    'version': '3.5.0',
    'description': 'Woke is a Python-based development and testing framework for Solidity.',
    'long_description': '# Woke\n\nWoke is a Python-based development and testing framework for Solidity.\n\nFeatures:\n\n- **Testing framework** - a testing framework for Solidity smart contracts with Python-native equivalents of Solidity types and blazing fast execution.\n\n- **Fuzzer** - a property-based fuzzer for Solidity smart contracts that allows testers to write their fuzz tests in Python.\n\n- **Vulnerability detectors**\n\n- **LSP server**\n\n## Dependencies\n\n- [Python](https://www.python.org/downloads/release/python-3910/) (version 3.7 or higher)\n\n> :warning: Python 3.11 is experimentally supported.\n\n## Installation\n\nvia `pip`\n\n```shell\npip3 install woke\n```\n\n## Documentation & Contribution\n\nWoke documentation can be found [here](https://ackeeblockchain.com/woke/docs/latest).\n\nThere you can also find a section on [contributing](https://ackeeblockchain.com/woke/docs/latest/contributing/).\n\n## Features\n\n### Testing framework\n\nSee [examples](examples) and [documentation](https://ackeeblockchain.com/woke/docs/latest/testing-framework/overview) for more information.\n\nWriting tests is as simple as:\n\n```python\nfrom woke.testing import *\nfrom pytypes.contracts.Counter import Counter\n\n@default_chain.connect()\ndef test_counter():\n    default_chain.set_default_accounts(default_chain.accounts[0])\n\n    counter = Counter.deploy()\n    assert counter.count() == 0\n\n    counter.increment()\n    assert counter.count() == 1\n```\n\n### Fuzzer\n\nFuzzer builds on top of the testing framework and allows efficient fuzz testing of Solidity smart contracts.\n\n```python\nfrom woke.testing import *\nfrom woke.testing.fuzzing import *\nfrom pytypes.contracts.Counter import Counter\n\nclass CounterTest(FuzzTest):\n    def pre_sequence(self) -> None:\n        self.counter = Counter.deploy()\n        self.count = 0\n\n    @flow()\n    def increment(self) -> None:\n        self.counter.increment()\n        self.count += 1\n\n    @flow()\n    def decrement(self) -> None:\n        with may_revert(Panic(PanicCodeEnum.UNDERFLOW_OVERFLOW)) as e:\n            self.counter.decrement()\n\n        if e.value is not None:\n            assert self.count == 0\n        else:\n            self.count -= 1\n\n    @invariant(period=10)\n    def count(self) -> None:\n        assert self.counter.count() == self.count\n\n@default_chain.connect()\ndef test_counter():\n    default_chain.set_default_accounts(default_chain.accounts[0])\n    CounterTest().run(sequences_count=30, flows_count=100)\n```\n\n### Vulnerability detectors\n\nVulnerability detectors can be run using:\n```shell\nwoke detect\n```\n\n### LSP server\n\nWoke implements an [LSP](https://microsoft.github.io/language-server-protocol/) server for Solidity. The only currently supported communication channel is TCP.\n\nWoke LSP server can be run using:\n\n```shell\nwoke lsp\n```\n\nOr with an optional --port argument:\n\n```shell\nwoke lsp --port 1234\n```\n\nAll LSP server features can be found in the [documentation](https://ackeeblockchain.com/woke/docs/latest/language-server/).\n\n## License\n\nThis project is licensed under the [ISC license](https://github.com/Ackee-Blockchain/woke/blob/main/LICENSE).\n\n## Partners\n\nRockawayX             |  Coinbase\n:-------------------------:|:-------------------------:\n[![](https://github.com/Ackee-Blockchain/woke/blob/main/images/rockawayx.jpg?raw=true)](https://rockawayx.com/)  |  [![](https://github.com/Ackee-Blockchain/woke/blob/main/images/coinbase.png?raw=true)](https://www.coinbase.com/)\n\n\n\n\n\n\n',
    'author': 'Ackee Blockchain',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://ackeeblockchain.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.9,<4.0.0',
}


setup(**setup_kwargs)



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

import woke.ast.ir.yul as yul
from woke.ast.enums import GlobalSymbolsEnum
from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.binary_operation import BinaryOperation
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.expression.unary_operation import UnaryOperation
from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.lsp.common_structures import (
    DocumentUri,
    Location,
    LocationLink,
    PartialResultParams,
    Position,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_compiler import LspCompiler
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path
from woke.lsp.utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class DefinitionOptions(WorkDoneProgressOptions):
    pass


class DefinitionRegistrationOptions(TextDocumentRegistrationOptions, DefinitionOptions):
    pass


class DefinitionParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    pass


def _create_location(
    path: Path, location: Tuple[int, int], compiler: LspCompiler
) -> Location:
    return Location(
        uri=DocumentUri(path_to_uri(path)),
        range=compiler.get_range_from_byte_offsets(path, location),
    )


def _get_results_from_node(
    original_node: Union[IrAbc, GlobalSymbolsEnum],
    position: Position,
    context: LspContext,
    byte_offset: int,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[List[Tuple[Path, Tuple[int, int]]]]:
    if isinstance(original_node, DeclarationAbc):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if position_within_range(position, name_location_range):
            return [(original_node.file, original_node.name_location)]
        return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(original_node, yul.Identifier):
        external_reference = original_node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration
    elif (
        isinstance(original_node, (UnaryOperation, BinaryOperation))
        and original_node.function is not None
    ):
        node = original_node.function
    else:
        node = original_node

    if not isinstance(node, (DeclarationAbc, SourceUnit)):
        return None

    definitions = []

    if isinstance(node, (FunctionDefinition, VariableDeclaration)):
        if isinstance(node, VariableDeclaration) or node.implemented:
            definitions.append((node.file, node.name_location))

        for base_function in node.base_functions:
            if base_function.implemented:
                definitions.append((base_function.file, base_function.name_location))
        if isinstance(node, FunctionDefinition):
            for child_function in node.child_functions:
                if isinstance(child_function, VariableDeclaration):
                    definitions.append(
                        (child_function.file, child_function.name_location)
                    )
                elif (
                    isinstance(child_function, FunctionDefinition)
                    and child_function.implemented
                ):
                    definitions.append(
                        (child_function.file, child_function.name_location)
                    )
    elif isinstance(node, ModifierDefinition):
        if node.implemented:
            definitions.append((node.file, node.name_location))
        for base_modifier in node.base_modifiers:
            if base_modifier.implemented:
                definitions.append((base_modifier.file, base_modifier.name_location))
        for child_modifier in node.child_modifiers:
            if child_modifier.implemented:
                definitions.append((child_modifier.file, child_modifier.name_location))
    elif isinstance(node, SourceUnit):
        definitions.append((node.file, node.byte_location))
    else:
        definitions.append((node.file, node.name_location))

    return definitions


async def _get_definition_from_cache(
    path: Path,
    position: Position,
    context: LspContext,
):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, (DeclarationAbc, SourceUnit)):
        if isinstance(node, DeclarationAbc):
            location = node.name_location
        else:
            location = node.byte_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, context, old_byte_offset, node_name_location
    )

    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        return _create_location(path, (new_start, new_end), context.compiler)
    else:
        ret = []
        for path, location in result:
            forward_changes = context.compiler.get_last_compilation_forward_changes(
                path
            )
            if forward_changes is None:
                raise Exception("No forward changes found")
            new_start = (
                changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
            )
            new_end = (
                changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
            )
            ret.append(_create_location(path, (new_start, new_end), context.compiler))
        return ret


async def definition(
    context: LspContext, params: DefinitionParams
) -> Union[Location, List[Location], List[LocationLink], None]:
    logger.debug(
        f"Requested definition for file {params.text_document.uri} at {params.position}"
    )

    path = uri_to_path(params.text_document.uri).resolve()
    if (
        path not in context.compiler.interval_trees
        or not context.compiler.output_ready.is_set()
    ):
        # try to use old build artifacts
        try:
            return await _get_definition_from_cache(path, params.position, context)
        except Exception:
            pass

    await context.compiler.output_ready.wait()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, DeclarationAbc):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, context, byte_offset, node_name_location
    )
    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        return _create_location(path, location, context.compiler)
    else:
        return [
            _create_location(path, location, context.compiler)
            for path, location in result
        ]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    SolcBlock,
    SolcBreak,
    SolcContinue,
    SolcDoWhileStatement,
    SolcEmitStatement,
    SolcExpressionStatement,
    SolcForStatement,
    SolcIfStatement,
    SolcInlineAssembly,
    SolcPlaceholderStatement,
    SolcReturn,
    SolcRevertStatement,
    SolcStatementUnion,
    SolcTryStatement,
    SolcUncheckedBlock,
    SolcVariableDeclarationStatement,
    SolcWhileStatement,
)

if TYPE_CHECKING:
    from ..declaration.function_definition import FunctionDefinition
    from ..declaration.modifier_definition import ModifierDefinition
    from ..meta.try_catch_clause import TryCatchClause
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class StatementAbc(SolidityAbc, ABC):
    """
    Abstract base class for all Solidity statements.
    """

    _documentation: Optional[str]

    def __init__(
        self, init: IrInitTuple, statement: SolcStatementUnion, parent: SolidityAbc
    ):
        super().__init__(init, statement, parent)
        self._documentation = statement.documentation

    @staticmethod
    def from_ast(
        init: IrInitTuple, statement: SolcStatementUnion, parent: SolidityAbc
    ) -> "StatementAbc":
        from .block import Block
        from .break_statement import Break
        from .continue_statement import Continue
        from .do_while_statement import DoWhileStatement
        from .emit_statement import EmitStatement
        from .expression_statement import ExpressionStatement
        from .for_statement import ForStatement
        from .if_statement import IfStatement
        from .inline_assembly import InlineAssembly
        from .placeholder_statement import PlaceholderStatement
        from .return_statement import Return
        from .revert_statement import RevertStatement
        from .try_statement import TryStatement
        from .unchecked_block import UncheckedBlock
        from .variable_declaration_statement import VariableDeclarationStatement
        from .while_statement import WhileStatement

        if isinstance(statement, SolcBlock):
            return Block(init, statement, parent)
        elif isinstance(statement, SolcBreak):
            return Break(init, statement, parent)
        elif isinstance(statement, SolcContinue):
            return Continue(init, statement, parent)
        elif isinstance(statement, SolcDoWhileStatement):
            return DoWhileStatement(init, statement, parent)
        elif isinstance(statement, SolcEmitStatement):
            return EmitStatement(init, statement, parent)
        elif isinstance(statement, SolcExpressionStatement):
            return ExpressionStatement(init, statement, parent)
        elif isinstance(statement, SolcForStatement):
            return ForStatement(init, statement, parent)
        elif isinstance(statement, SolcIfStatement):
            return IfStatement(init, statement, parent)
        elif isinstance(statement, SolcInlineAssembly):
            return InlineAssembly(init, statement, parent)
        elif isinstance(statement, SolcPlaceholderStatement):
            return PlaceholderStatement(init, statement, parent)
        elif isinstance(statement, SolcReturn):
            return Return(init, statement, parent)
        elif isinstance(statement, SolcRevertStatement):
            return RevertStatement(init, statement, parent)
        elif isinstance(statement, SolcTryStatement):
            return TryStatement(init, statement, parent)
        elif isinstance(statement, SolcUncheckedBlock):
            return UncheckedBlock(init, statement, parent)
        elif isinstance(statement, SolcVariableDeclarationStatement):
            return VariableDeclarationStatement(init, statement, parent)
        elif isinstance(statement, SolcWhileStatement):
            return WhileStatement(init, statement, parent)
        assert False, f"Unknown statement type: {type(statement)}"

    @property
    @abstractmethod
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
        FunctionDefinition,
        ModifierDefinition,
        TryCatchClause,
    ]:
        """
        Returns:
            Parent node of the statement.
        """
        ...

    @property
    @abstractmethod
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        """
        Returns:
            Set of child IR nodes (including `self`) that modify the blockchain state and flags describing how the state is modified.
        """
        ...

    def statements_iter(self) -> Iterator[StatementAbc]:
        """
        Yields:
            Child statements of the statement (recursively) including `self`.
        """
        yield self

    @property
    def documentation(self) -> Optional[str]:
        """
        Statement documentation strings should be placed above the statement.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation



import logging

import graphviz as gv

from woke.config.data_model import ImportsDirection
from woke.lsp.context import LspContext

logger = logging.getLogger(__name__)


async def generate_imports_graph_handler(context: LspContext) -> str:
    logger.debug(f"Requested imports graph")

    await context.compiler.output_ready.wait()

    config = context.config.generator.imports_graph

    g = gv.Digraph("Imports graph")
    g.attr(rankdir=config.direction)
    g.attr("node", shape="box")

    graph = context.compiler.last_graph

    for node in graph.nodes:  # pyright: ignore reportGeneralTypeIssues
        node_attrs = {}
        if config.vscode_urls:
            node_attrs[
                "URL"
            ] = f"vscode://file/{graph.nodes[node]['path']}"  # pyright: ignore reportGeneralTypeIssues
        g.node(node, **node_attrs)

    for from_, to in graph.edges:  # pyright: ignore reportGeneralTypeIssues
        if config.imports_direction == ImportsDirection.ImportedToImporting:
            g.edge(from_, to)
        else:
            g.edge(to, from_)

    return g.source



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Optional

from ..config import WokeConfig
from ..core.solidity_version import SemanticVersion
from ..utils.openzeppelin import get_contracts_package_version
from .features.diagnostic import diagnostics_loop
from .lsp_compiler import LspCompiler
from .lsp_parser import LspParser

if TYPE_CHECKING:
    from .server import LspServer


class LspContext:
    __server: LspServer
    __workspace_config: WokeConfig
    __compiler: LspCompiler
    __diagnostics_queue: asyncio.Queue
    __openzeppelin_contracts_version: Optional[SemanticVersion]
    __parser: LspParser

    def __init__(
        self, server: LspServer, config: WokeConfig, perform_files_discovery: bool
    ) -> None:
        self.__server = server
        self.__workspace_config = config
        self.__diagnostics_queue = asyncio.Queue()
        self.__compiler = LspCompiler(
            server, self.__diagnostics_queue, perform_files_discovery
        )
        self.__openzeppelin_contracts_version = get_contracts_package_version(config)
        self.__parser = LspParser(server)

    def run(self) -> None:
        self.__server.create_task(diagnostics_loop(self.__server, self))
        self.__server.create_task(self.__compiler.run(self.__workspace_config))

    @property
    def config(self) -> WokeConfig:
        return self.__workspace_config

    @property
    def compiler(self) -> LspCompiler:
        return self.__compiler

    @property
    def diagnostics_queue(self) -> asyncio.Queue:
        return self.__diagnostics_queue

    @property
    def server(self) -> LspServer:
        return self.__server

    @property
    def openzeppelin_contracts_version(self) -> Optional[SemanticVersion]:
        return self.__openzeppelin_contracts_version

    @property
    def parser(self) -> LspParser:
        return self.__parser
